#!/opt/conda/bin/python
# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import argparse
import os
import sys
from typing import List

import urllib3
import yaml
from jinja2 import Environment, FileSystemLoader, select_autoescape
from kubernetes import client, config
from kubernetes.client.rest import ApiException

urllib3.disable_warnings()

KERNEL_POD_TEMPLATE_PATH = "/kernel-pod.yaml.j2"


def generate_kernel_pod_yaml(keywords):
    """Return the kubernetes pod spec as a yaml string.

    - load jinja2 template from this file directory.
    - substitute template variables with keywords items.
    """
    j_env = Environment(
        loader=FileSystemLoader(os.path.dirname(__file__)),
        trim_blocks=True,
        lstrip_blocks=True,
        autoescape=select_autoescape(
            disabled_extensions=(
                "j2",
                "yaml",
            ),
            default_for_string=True,
            default=True,
        ),
    )
    # jinja2 template substitutes template variables with None though keywords doesn't
    # contain corresponding item. Therefore, no need to check if any are left unsubstituted.
    # Kubernetes API server will validate the pod spec instead.
    k8s_yaml = j_env.get_template(KERNEL_POD_TEMPLATE_PATH).render(**keywords)

    return k8s_yaml


def extend_pod_env(pod_def: dict) -> dict:
    """Extends the pod_def.spec.containers[0].env stanza with current environment."""
    env_stanza = pod_def["spec"]["containers"][0].get("env") or []

    # Walk current set of template env entries and replace those found in the current
    # env with their values (and record those items).   Then add all others from the env
    # that were not already.
    processed_entries: List[str] = []
    for item in env_stanza:
        item_name = item.get("name")
        if item_name in os.environ:
            item["value"] = os.environ[item_name]
            processed_entries.append(item_name)

    for name, value in os.environ.items():
        if name not in processed_entries:
            env_stanza.append({"name": name, "value": value})

    pod_def["spec"]["containers"][0]["env"] = env_stanza
    return pod_def


# A popular reason that lasts many APIs but is not a constant in the client lib
K8S_ALREADY_EXIST_REASON = "AlreadyExists"


def _parse_k8s_exception(exc: ApiException) -> str:
    """Parse the exception and return the error message from kubernetes api
    Args:
        exc (Exception): Exception object
    Returns:
        str: Error message from kubernetes api
    """
    # more exception can be parsed, but at the time of implementation we only need this one
    if exc.status == 409:
        if exc.reason == "Conflict" and f'"reason":{K8S_ALREADY_EXIST_REASON}' in exc.body:
            return K8S_ALREADY_EXIST_REASON
    return ""


def launch_kubernetes_kernel(
    kernel_id,
    port_range,
    response_addr,
    public_key,
    spark_context_init_mode,
    pod_template_file,
    spark_opts_out,
    kernel_class_name,
):
    # Launches a containerized kernel as a kubernetes pod.

    if os.getenv("KUBERNETES_SERVICE_HOST"):
        config.load_incluster_config()
    else:
        config.load_kube_config()

    # Capture keywords and their values.
    keywords = {}

    # Factory values...
    # Since jupyter lower cases the kernel directory as the kernel-name, we need to capture its case-sensitive
    # value since this is used to locate the kernel launch script within the image.
    # Ensure these key/value pairs are reflected in the environment.  We'll add these to the container's env
    # stanza after the pod template is generated.
    if port_range:
        os.environ["PORT_RANGE"] = port_range
    if public_key:
        os.environ["PUBLIC_KEY"] = public_key
    if response_addr:
        os.environ["RESPONSE_ADDRESS"] = response_addr
    if kernel_id:
        os.environ["KERNEL_ID"] = kernel_id
    if spark_context_init_mode:
        os.environ["KERNEL_SPARK_CONTEXT_INIT_MODE"] = spark_context_init_mode
    if kernel_class_name:
        os.environ["KERNEL_CLASS_NAME"] = kernel_class_name

    os.environ["KERNEL_NAME"] = os.path.basename(
        os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    )

    # Walk env variables looking for names prefixed with KERNEL_.  When found, set corresponding keyword value
    # with name in lower case.
    for name, value in os.environ.items():
        if name.startswith("KERNEL_"):
            keywords[name.lower()] = yaml.safe_load(value)

    # Substitute all template variable (wrapped with {{ }}) and generate `yaml` string.
    k8s_yaml = generate_kernel_pod_yaml(keywords)

    # For each k8s object (kind), call the appropriate API method.  Too bad there isn't a method
    # that can take a set of objects.
    #
    # Creation for additional kinds of k8s objects can be added below.  Refer to
    # https://github.com/kubernetes-client/python for API signatures.  Other examples can be found in
    # https://github.com/jupyter-server/enterprise_gateway/tree/main/enterprise_gateway/services/processproxies/k8s.py
    #
    pod_template = None
    kernel_namespace = keywords["kernel_namespace"]
    k8s_objs = yaml.safe_load_all(k8s_yaml)
    for k8s_obj in k8s_objs:
        if k8s_obj.get("kind"):
            if k8s_obj["kind"] == "Pod":
                #  print("{}".format(k8s_obj))  # useful for debug
                pod_template = extend_pod_env(k8s_obj)
                if pod_template_file is None:
                    try:
                        client.CoreV1Api(client.ApiClient()).create_namespaced_pod(
                            body=k8s_obj, namespace=kernel_namespace
                        )
                    except ApiException as exc:
                        if _parse_k8s_exception(exc) == K8S_ALREADY_EXIST_REASON:
                            (
                                client.CoreV1Api(client.ApiClient())
                                .list_namespaced_pod(
                                    namespace=kernel_namespace,
                                    label_selector=f"kernel_id={kernel_id}",
                                    watch=False,
                                )
                                .items[0]
                            )
                        else:
                            raise exc
            elif k8s_obj["kind"] == "Secret":
                if pod_template_file is None:
                    client.CoreV1Api(client.ApiClient()).create_namespaced_secret(
                        body=k8s_obj, namespace=kernel_namespace
                    )
            elif k8s_obj["kind"] == "PersistentVolumeClaim":
                if pod_template_file is None:
                    try:
                        client.CoreV1Api(
                            client.ApiClient()
                        ).create_namespaced_persistent_volume_claim(
                            body=k8s_obj, namespace=kernel_namespace
                        )
                    except ApiException as exc:
                        if _parse_k8s_exception(exc) == K8S_ALREADY_EXIST_REASON:
                            pass
                        else:
                            raise exc
            elif k8s_obj["kind"] == "PersistentVolume":
                if pod_template_file is None:
                    client.CoreV1Api(client.ApiClient()).create_persistent_volume(body=k8s_obj)
            else:
                sys.exit(
                    f"ERROR - Unhandled Kubernetes object kind '{k8s_obj['kind']}' found in yaml file - "
                    f"kernel launch terminating!"
                )
        else:
            print("ERROR processing Kubernetes yaml file - kernel launch terminating!")
            print(k8s_yaml)
            sys.exit(
                f"ERROR - Unknown Kubernetes object '{k8s_obj}' found in yaml file - kernel launch terminating!"
            )

    if pod_template_file:
        # TODO - construct other --conf options for things like mounts, resources, etc.
        # write yaml to file...
        stream = open(pod_template_file, "w")
        yaml.dump(pod_template, stream)

        # Build up additional spark options.  Note the trailing space to accommodate concatenation
        additional_spark_opts = (
            f"--conf spark.kubernetes.driver.podTemplateFile={pod_template_file} "
            f"--conf spark.kubernetes.executor.podTemplateFile={pod_template_file} "
        )

        additional_spark_opts += _get_spark_resources(pod_template)

        if spark_opts_out:
            with open(spark_opts_out, "w+") as soo_fd:
                soo_fd.write(additional_spark_opts)
        else:  # If no spark_opts_out was specified, print to stdout in case this is an old caller
            print(additional_spark_opts)


def _get_spark_resources(pod_template: dict) -> str:
    # Gather up resources for cpu/memory requests/limits.  Since gpus require a "discovery script"
    # we'll leave that alone for now:
    # https://spark.apache.org/docs/latest/running-on-kubernetes.html#resource-allocation-and-configuration-overview
    #
    # The config value names below are pulled from:
    # https://spark.apache.org/docs/latest/running-on-kubernetes.html#container-spec
    spark_resources = ""
    containers = pod_template.get("spec", {}).get("containers", [])
    if containers:
        # We're just dealing with single-container pods at this time.
        resources = containers[0].get("resources", {})
        if resources:
            requests = resources.get("requests", {})
            if requests:
                cpu_request = requests.get("cpu")
                if cpu_request:
                    spark_resources += (
                        f"--conf spark.driver.cores={cpu_request} "
                        f"--conf spark.executor.cores={cpu_request} "
                    )
                memory_request = requests.get("memory")
                if memory_request:
                    spark_resources += (
                        f"--conf spark.driver.memory={memory_request} "
                        f"--conf spark.executor.memory={memory_request} "
                    )

            limits = resources.get("limits", {})
            if limits:
                cpu_limit = limits.get("cpu")
                if cpu_limit:
                    spark_resources += (
                        f"--conf spark.kubernetes.driver.limit.cores={cpu_limit} "
                        f"--conf spark.kubernetes.executor.limit.cores={cpu_limit} "
                    )
                memory_limit = limits.get("memory")
                if memory_limit:
                    spark_resources += (
                        f"--conf spark.driver.memory={memory_limit} "
                        f"--conf spark.executor.memory={memory_limit} "
                    )
    return spark_resources


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--kernel-id",
        dest="kernel_id",
        nargs="?",
        help="Indicates the id associated with the launched kernel.",
    )
    parser.add_argument(
        "--port-range",
        dest="port_range",
        nargs="?",
        metavar="<lowerPort>..<upperPort>",
        help="Port range to impose for kernel ports",
    )
    parser.add_argument(
        "--response-address",
        dest="response_address",
        nargs="?",
        metavar="<ip>:<port>",
        help="Connection address (<ip>:<port>) for returning connection file",
    )
    parser.add_argument(
        "--public-key",
        dest="public_key",
        nargs="?",
        help="Public key used to encrypt connection information",
    )
    parser.add_argument(
        "--spark-context-initialization-mode",
        dest="spark_context_init_mode",
        nargs="?",
        help="Indicates whether or how a spark context should be created",
        default="none",
    )
    parser.add_argument(
        "--pod-template",
        dest="pod_template_file",
        nargs="?",
        metavar="template filename",
        help="When present, yaml is written to file, no launch performed.",
    )
    parser.add_argument(
        "--spark-opts-out",
        dest="spark_opts_out",
        nargs="?",
        metavar="additional spark options filename",
        help="When present, additional spark options are written to file, "
        "no launch performed, requires --pod-template.",
    )
    parser.add_argument(
        "--kernel-class-name",
        dest="kernel_class_name",
        nargs="?",
        help="Indicates the name of the kernel class to use.  Must be a subclass of 'ipykernel.kernelbase.Kernel'.",
    )

    arguments = vars(parser.parse_args())
    kernel_id = arguments["kernel_id"]
    port_range = arguments["port_range"]
    response_addr = arguments["response_address"]
    public_key = arguments["public_key"]
    spark_context_init_mode = arguments["spark_context_init_mode"]
    pod_template_file = arguments["pod_template_file"]
    spark_opts_out = arguments["spark_opts_out"]
    kernel_class_name = arguments["kernel_class_name"]

    launch_kubernetes_kernel(
        kernel_id,
        port_range,
        response_addr,
        public_key,
        spark_context_init_mode,
        pod_template_file,
        spark_opts_out,
        kernel_class_name,
    )



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
import os

from jupyter_core.application import JupyterApp
from overrides import overrides
from traitlets import Bool, Unicode, default

from .._version import __version__
from .base_app import DEFAULT_LANGUAGE, PYTHON, SCALA, BaseSpecApp, R

DEFAULT_KERNEL_NAMES = {PYTHON: "docker_python", SCALA: "docker_scala", R: "docker_r"}
KERNEL_SPEC_TEMPLATE_NAMES = {
    PYTHON: "container_python",
    SCALA: "container_scala",
    R: "container_r",
}
DEFAULT_DISPLAY_NAMES = {PYTHON: "Docker Python", SCALA: "Docker Scala", R: "Docker R"}
DEFAULT_IMAGE_NAMES = {
    PYTHON: "elyra/kernel-py",
    SCALA: "elyra/kernel-scala",
    R: "elyra/kernel-r",
}

DOCKER_PROVISIONER_NAME = "docker-provisioner"
SWARM_PROVISIONER_NAME = "docker-swarm-provisioner"
LAUNCHER_NAME = "launch_docker.py"


class DockerSpecInstaller(BaseSpecApp):
    """CLI for extension management."""

    name = "jupyter-docker-spec"
    description = (
        "Creates a Jupyter kernel specification for use within a Docker or Docker Swarm cluster."
    )
    # Note that the left justification of the second example is necessary to ensure proper
    # alignment with the first example during --help output.
    examples = """
    jupyter-docker-spec install --language=R --kernel-name=r_docker --image_name=foo/my_r_kernel_image:v4_0

jupyter-docker-spec install --swarm --kernel-name=python_swarm
    """

    @default("kernel_name")
    def _kernel_name_default(self) -> str:
        return DEFAULT_KERNEL_NAMES[DEFAULT_LANGUAGE]

    @default("display_name")
    def _display_name_default(self) -> str:
        return DEFAULT_DISPLAY_NAMES[DEFAULT_LANGUAGE]

    # Image name
    image_name_env = "GP_IMAGE_NAME"
    image_name = Unicode(
        None,
        config=True,
        allow_none=True,
        help="""The kernel image to use for this kernel specification. If this specification is
enabled for Spark usage, this image will be the driver image. (GP_IMAGE_NAME env var)""",
    )

    @default("image_name")
    def _image_name_default(self):
        return os.getenv(self.image_name_env)

    # Flags
    swarm = Bool(
        False, config=True, help="Install kernelspec for use within a Docker Swarm cluster."
    )

    provisioner_name = Unicode(DOCKER_PROVISIONER_NAME, config=False)
    launcher_name = Unicode(LAUNCHER_NAME, config=False)

    aliases = {
        "image-name": "DockerSpecInstaller.image_name",
    }
    aliases.update(BaseSpecApp.aliases)

    flags = {
        "swarm": (
            {"DockerSpecInstaller": {"swarm": True}},
            swarm.help,
        ),
    }
    flags.update(BaseSpecApp.flags)

    @overrides
    def detect_missing_extras(self):
        super().detect_missing_extras()
        try:
            import docker  # noqa: F401
        except ImportError:
            self.log.warning(
                "The extra package 'docker' is not installed in this environment and is required.  "
                "Ensure that gateway_provisioners is installed by specifying the extra 'docker' "
                "(e.g., pip install 'gateway_provisioners[docker]')."
            )

    @overrides
    def validate_parameters(self):
        super().validate_parameters()

        self.language = self.language.lower()
        self.launcher_dir_name = "docker"
        self.resource_dir_name = self.language

        self.provisioner_name = SWARM_PROVISIONER_NAME if self.swarm else DOCKER_PROVISIONER_NAME

        if self.kernel_name == DEFAULT_KERNEL_NAMES[DEFAULT_LANGUAGE]:
            self.kernel_name = DEFAULT_KERNEL_NAMES[self.language]
        if self.display_name == DEFAULT_DISPLAY_NAMES[DEFAULT_LANGUAGE]:
            self.display_name = DEFAULT_DISPLAY_NAMES[self.language]

        self.kernel_spec_dir_name = KERNEL_SPEC_TEMPLATE_NAMES[self.language]

        if self.image_name is None:
            self.image_name = f"{DEFAULT_IMAGE_NAMES[self.language]}:{self._get_tag()}"

        # sanitize kernel_name
        self.kernel_name = self.kernel_name.replace(" ", "_")

    @overrides
    def get_substitutions(self, install_dir) -> dict:
        substitutions = super().get_substitutions(install_dir)
        substitutions["image_name"] = self.image_name
        substitutions["provisioner_name"] = self.provisioner_name
        substitutions["launcher_name"] = self.launcher_name
        return substitutions


class DockerProvisionerApp(JupyterApp):
    """Application responsible for driving the creation of Docker-based kernel specifications."""

    version = __version__
    name = "jupyter docker-spec"
    description = """Application used to create kernel specifications for use on Docker or Docker Swarm clusters
    via the DockerProvisioner or DockerSwarmProvisioner kernel provisioners."""
    subcommands = dict(
        {
            "install": (
                DockerSpecInstaller,
                DockerSpecInstaller.description.splitlines()[0],
            ),
        }
    )
    aliases = {}
    flags = {}

    def start(self):
        super().start()

        if self.subapp is None:
            print(f"No subcommand specified. Must specify one of: {list(self.subcommands)}")
            print()
            self.print_description()
            self.print_subcommands()
            self.exit(1)


if __name__ == "__main__":
    DockerProvisionerApp.launch_instance()



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
from __future__ import annotations

import json
import os
import shutil
import sys
import tempfile
from glob import glob
from string import Template
from typing import Any

from jupyter_client.kernelspec import KernelSpec, KernelSpecManager
from jupyter_core.application import JupyterApp, base_aliases, base_flags
from jupyter_core.paths import SYSTEM_JUPYTER_PATH, jupyter_data_dir
from overrides import overrides
from traitlets import Bool, Instance, TraitError, Unicode, default, validate

from .._version import __version__
from ..config_mixin import RemoteProvisionerConfigMixin

kernel_launchers_dir = os.path.join(os.path.dirname(__file__), "..", "kernel-launchers")
kernel_resources_dir = os.path.join(os.path.dirname(__file__), "..", "kernel-resources")
kernel_specs_dir = os.path.join(os.path.dirname(__file__), "..", "kernel-specs")

launcher_dirs = ["python", "r", "scala", "kubernetes", "docker", "operators"]
resource_dirs = ["python", "r", "scala"]

KERNEL_JSON = "kernel.json"
PYTHON = "python"
SCALA = "scala"
R = "r"
DEFAULT_LANGUAGE = PYTHON
SUPPORTED_LANGUAGES = [PYTHON, SCALA, R]
DEFAULT_INIT_MODE = "lazy"
SPARK_INIT_MODES = [DEFAULT_INIT_MODE, "eager", "none"]
LANGUAGE_SUBSTITUTIONS = {PYTHON: PYTHON, R: "R", SCALA: SCALA}
DEFAULT_PYTHON_KERNEL_CLASS_NAME = "ipykernel.ipkernel.IPythonKernel"


class BaseApp(JupyterApp):
    """Base class containing parameters common to each provisioner."""

    kernel_spec_manager = Instance(KernelSpecManager)

    @default("kernel_spec_manager")
    def _kernel_spec_manager_default(self) -> KernelSpecManager:
        return KernelSpecManager()

    launcher_dir_name = Unicode()  # kernel-launchers directory name
    resource_dir_name = Unicode()  # kernel-resources directory name
    kernel_spec_dir_name = Unicode()  # kernel-specs directory name
    install_dir = Unicode()  # Final location for kernel spec
    toree_jar_path: str | None = None  # Location from which to copy the toree jar

    kernel_spec_install: bool = True  # Set to false when bootstrap installation occurs

    def start(self):
        """Drive the kernel specification creation."""
        super().start()

        self.validate_parameters()
        self.detect_missing_extras()
        self.install_files()

    def validate_parameters(self):
        """
        Validate input parameters and prepare for their injection into templated files.

        This method is overridden by subclasses which should call super().validate_parameters().
        """
        pass

    def detect_missing_extras(self):
        """
        Issues a warning message whenever an "extra" library is detected as missing.

        Note that "extra" can also mean things like Apache Toree is not installed when
        the language is Scala, or Rscript is not available when the language is R.
        """
        pass

    def _detect_missing_toree_jar(self):
        """
        Detects which aspects of Apache Toree are missing.

        If installed, then it determines the path to the toree jar file.  If the jar cannot be
        determined, appropriate warnings are issued.
        """
        self.toree_jar_path = None
        try:
            import toree
        except ImportError:
            self.log.warning(
                "The Apache Torre kernel package is not installed in this environment and is required "
                "for kernels of language 'Scala'.  Ensure that the 'apache-toree' package is installed "
                "(e.g., pip install 'apache-toree') then repeat this command to ensure the Apache Toree"
                "jar file is located in the kernel specification's lib directory prior to its use."
            )
        else:
            toree_version = toree.toreeapp.ToreeApp.version
            toree_lib_dir = os.path.join(os.path.dirname(toree.__file__), "lib")
            jars = glob(os.path.join(toree_lib_dir, f"toree-assembly-{toree_version}-*.jar"))
            if len(jars) < 1:
                self.log.warning(
                    "The Apache Torre kernel package is installed, but there doesn't appear to be a toree "
                    f"jar file located in the installation area: '{toree_lib_dir}' that matches the pattern "
                    f"'toree-assembly-{toree_version}-*.jar'.  This jar file is required for the proper "
                    "behavior of scala kernels."
                )
            elif len(jars) > 1:
                self.log.warning(
                    "The Apache Torre kernel package is installed, but there appears to be more than one "
                    f"toree-assembly jar file located in the installation area: '{toree_lib_dir}' that matches "
                    f"the pattern 'toree-assembly-{toree_version}-*.jar'.  You will need to ensure the appropriate "
                    "jar file is copied to the kernel specification's lib directory prior to its use."
                )
            else:
                self.toree_jar_path = jars[0]

    def _detect_missing_rscript(self):
        """Detects if Rscript is in current path and issues warning if not."""
        rscript_location = shutil.which("Rscript")
        if rscript_location is None:
            self.log.warning(
                "The executable 'Rscript' is not in the current PATH.  Please ensure that 'Rscript' "
                "is available prior to using this kernel specification."
            )

    def install_files(self):
        """Installs applicable files into the target location."""
        pass

    def _copy_launcher_files(self, launcher_dir_name: str, target_dir: str):
        """Copy the launcher files from the launcher directory to the destination staging directory."""

        src_dir = os.path.join(kernel_launchers_dir, launcher_dir_name)
        shutil.copytree(src=src_dir, dst=target_dir, dirs_exist_ok=True)

        # When the launcher_dir_name is either 'r' or 'python', we need to also copy the files
        # from the 'shared' launcher directory.
        if launcher_dir_name.lower() in [PYTHON, R]:
            src_dir = os.path.join(kernel_launchers_dir, "shared")
            shutil.copytree(src=src_dir, dst=target_dir, dirs_exist_ok=True)
        # When the launcher_dir_name is 'scala', we need to copy the toree jar (if determined), and
        # remove the toree-launcher source code from the staging dir.
        if launcher_dir_name in [SCALA]:
            if self.toree_jar_path:
                shutil.copyfile(
                    self.toree_jar_path,
                    os.path.join(target_dir, "lib", os.path.basename(self.toree_jar_path)),
                )
            scala_src_dir = os.path.join(target_dir, "toree-launcher")
            if os.path.isdir(scala_src_dir):
                self._delete_directory(scala_src_dir)

        # The source launcher 'scripts' directory may contain a __pycache__ directory.
        # Check for this condition in the staging area and delete the directory if present.
        pycache_dir = os.path.join(target_dir, "scripts", "__pycache__")
        if os.path.isdir(pycache_dir):
            self._delete_directory(pycache_dir)

    def log_and_exit(self, msg, exit_status=1):
        """Logs the msg as an error and exits with the given exit-status."""
        self.log.error(msg)
        self.exit(exit_status)

    @staticmethod
    def _create_staging_directory(parent_dir=None):
        """Creates a temporary staging directory at the specified location.

        If no `parent_dir` is specified, the platform-specific "temp" directory is used.
        """
        return tempfile.mkdtemp(prefix="staging_", dir=parent_dir)

    @staticmethod
    def _delete_directory(dir_name):
        """Deletes the specified directory."""
        shutil.rmtree(dir_name)

    @staticmethod
    def import_item(name: str) -> Any:
        """
        Import and return ``bar`` given the string ``foo.bar``.
        Calling ``bar = import_item("foo.bar")`` is the functional equivalent of
        executing the code ``from foo import bar``.

        :param name : string
          The fully qualified name of the module/package being imported.

        :returns mod : module object
           The module that was imported.
        """

        parts = name.rsplit(".", 1)
        if len(parts) == 2:
            # called with 'foo.bar....'
            package, obj = parts
            module = __import__(package, fromlist=[obj])
            try:
                pak = getattr(module, obj)
            except AttributeError as ae:
                err_msg = f"No module named '{obj}'"
                raise ImportError(err_msg) from ae
            return pak
        else:
            # called with un-dotted string
            return __import__(parts[0])

    @staticmethod
    def _get_tag() -> str:
        """Determines the tag to use for images based on the version.

        If the version indicates a development version, the tag will be "dev",
        else the tag will represent the version, including pre-releases like 2.0.0rc1
        """
        if "dev" in __version__:
            return "dev"
        return __version__


class BaseSpecApp(RemoteProvisionerConfigMixin, BaseApp):
    kernel_name = Unicode(
        config=True, help="""Install the kernel spec into a directory with this name."""
    )

    display_name = Unicode(
        config=True,
        help="""The display name of the kernel - used by user-facing applications.""",
    )

    language = Unicode(
        "Python",
        config=True,
        help="""The language of the kernel referenced in the kernel specification.  Must be one of
    'Python', 'R', or 'Scala'.  Default = 'Python'.""",
    )

    @validate("language")
    def _language_validate(self, proposal: dict[str, str]) -> str:
        value = proposal["value"]
        try:
            assert value.lower() in SUPPORTED_LANGUAGES
        except AssertionError as ae:
            err_msg = f"Invalid language value {value}, not in {SUPPORTED_LANGUAGES}"
            raise TraitError(err_msg) from ae
        return value

    ipykernel_subclass_name = Unicode(
        DEFAULT_PYTHON_KERNEL_CLASS_NAME,
        config=True,
        help="""For Python kernels, the name of the ipykernel subclass.""",
    )
    # Flags
    user = Bool(
        False,
        config=True,
        help="Try to install the kernel spec to the per-user directory instead of the system "
        "or environment directory.",
    )

    prefix = Unicode(
        "",
        config=True,
        help="Specify a prefix to install to, e.g. an env. The kernelspec will be "
        "installed in PREFIX/share/jupyter/kernels/",
    )

    replace = Bool(
        False,
        config=True,
        help="If a kernel specification already exists in the destination, allow for its replacement.",
    )

    aliases = {
        "prefix": "BaseSpecApp.prefix",
        "kernel-name": "BaseSpecApp.kernel_name",
        "display-name": "BaseSpecApp.display_name",
        "language": "BaseSpecApp.language",
        "authorized-users": "BaseSpecApp.authorized_users",
        "unauthorized-users": "BaseSpecApp.unauthorized_users",
        "port-range": "BaseSpecApp.port_range",
        "launch-timeout": "BaseSpecApp.launch_timeout",
        "ipykernel-subclass-name": "BaseSpecApp.ipykernel_subclass_name",
    }
    aliases.update(base_aliases)

    flags = {
        "user": (
            {"BaseSpecApp": {"user": True}},
            user.help,
        ),
        "replace": (
            {"BaseSpecApp": {"replace": True}},
            replace.help,
        ),
        "sys-prefix": (
            {"BaseSpecApp": {"prefix": sys.prefix}},
            prefix.help,
        ),
        "debug": base_flags["debug"],
    }

    @overrides
    def validate_parameters(self):
        if self.user and self.prefix:
            self.log_and_exit("Can't specify both user and prefix. Please choose one or the other.")

        if self.ipykernel_subclass_name != DEFAULT_PYTHON_KERNEL_CLASS_NAME:
            if self.language.lower() != PYTHON:
                self.log.warning(
                    "--ipykernel_subclass_name will be ignored since --language is not Python."
                )
            else:  # Attempt to validate the value is a subclass of ipykernel, but only warn if ImportError
                try:
                    from ipykernel.ipkernel import IPythonKernel

                    ipykernel_subclass = self.import_item(self.ipykernel_subclass_name)
                    if not issubclass(ipykernel_subclass, IPythonKernel):
                        self.log_and_exit(
                            f"Parameter ipykernel-subclass-name of '{self.ipykernel_subclass_name}' "
                            f"does not appear to be a subclass of '{DEFAULT_PYTHON_KERNEL_CLASS_NAME}'"
                        )
                except ImportError:
                    self.log.warning(
                        f"Cannot determine if parameter ipykernel-subclass-name "
                        f"'{self.ipykernel_subclass_name}' is a subclass of "
                        f"'{DEFAULT_PYTHON_KERNEL_CLASS_NAME}'.  Continuing..."
                    )

    @overrides
    def detect_missing_extras(self):
        if self.launcher_dir_name in [SCALA]:
            self._detect_missing_toree_jar()

        if self.launcher_dir_name in [R]:
            self._detect_missing_rscript()

    @overrides
    def install_files(self):
        """Assembles kernel-specs, launchers and resources into staging directory, then installs as kernel-spec."""

        # Before setting up staging area, check if kernel spec already exists and no replacement has been requested
        dest_dir, temp_dir = self._replace_existing()

        # create staging dir
        staging_dir = self._create_staging_directory()

        # copy appropriate resource files
        self._copy_kernel_spec_files(staging_dir)

        try:
            # install to destination
            self.log.info(f"Installing kernel specification for '{self.display_name}'")
            self.install_dir = self.kernel_spec_manager.install_kernel_spec(
                staging_dir,
                kernel_name=self.kernel_name,
                user=self.user,
                prefix=self.prefix,
            )
            self._delete_directory(staging_dir)
            # If we're installing a scala kernel and don't have the toree jar file, issue
            # a warning indicating that the scala kernel needs that file in its kernelspec
            # directory hierarchy.
            if self.launcher_dir_name in [SCALA] and not self.toree_jar_path:
                kspec_toree_jar_location = os.path.join(self.install_dir, "lib")
                self.log.warning(
                    "The Apache Toree kernel is either not installed or it's jar file cannot be determined. "
                    f"Please ensure that the Toree jar file is placed into '{kspec_toree_jar_location}' "
                    f"prior to using the kernel."
                )
            # Apply substitutions to kernel.json file
            self._finalize_kernel_json()

        except Exception as ex:
            # We encountered an exception.  If we're in replace mode, revert temp directory back to destination
            if temp_dir:
                shutil.rmtree(dest_dir)
                shutil.copytree(src=temp_dir, dst=dest_dir, dirs_exist_ok=True)
                self.log.warning(
                    f"An exception was encountered while finalizing the kernel specification and "
                    f"the previous contents have been restored. The exception was: {ex}"
                )
        finally:
            if temp_dir:
                shutil.rmtree(temp_dir)

    def _copy_kernel_spec_files(self, staging_dir: str):
        """Copies the launcher, resource and kernel-spec files to the staging directory."""

        if any(
            dir_name is None
            for dir_name in [
                self.launcher_dir_name,
                self.resource_dir_name,
                self.kernel_spec_dir_name,
            ]
        ):
            err_msg = (
                "Invalid parameters.  Each of launcher_dir_name, resource_dir_name, "
                "and kernel_spec_dir_name must have a value!"
            )
            raise ValueError(err_msg)

        if self.launcher_dir_name not in launcher_dirs:
            err_msg = (
                f"Invalid launcher_dir_name '{self.launcher_dir_name}' "
                f"detected! Must be one of: {launcher_dirs}"
            )
            raise ValueError(err_msg)

        if self.resource_dir_name not in resource_dirs:
            err_msg = (
                f"Invalid resource_dir_name '{self.resource_dir_name}' "
                f"detected! Must be one of: {resource_dirs}"
            )
            raise ValueError(err_msg)

        # Copy the launcher files
        self._copy_launcher_files(self.launcher_dir_name, staging_dir)

        # Copy the resource files
        src_dir = os.path.join(kernel_resources_dir, self.resource_dir_name)
        shutil.copytree(src=src_dir, dst=staging_dir, dirs_exist_ok=True)

        # Copy the kernel-spec files
        src_dir = os.path.join(kernel_specs_dir, self.kernel_spec_dir_name)
        shutil.copytree(src=src_dir, dst=staging_dir, dirs_exist_ok=True)

    def _finalize_kernel_json(self):
        """Apply substitutions to the kernel.json string, update a kernel spec using these values,
        then write to the target kernel.json file.
        """
        subs = self.get_substitutions(self.install_dir)
        kernel_json_str = ""
        with open(os.path.join(self.install_dir, KERNEL_JSON)) as f:
            for line in f:
                line = line.split("#", 1)[0]
                kernel_json_str = kernel_json_str + line
            f.close()
        post_subs = Template(kernel_json_str).safe_substitute(subs)
        kernel_json = json.loads(post_subs)

        # Instantiate default KernelSpec, then update with the substitutions.  This allows for new fields
        # to be added that we might not yet know about.
        kernel_spec = KernelSpec().to_dict()
        kernel_spec.update(kernel_json)

        # Add ad-hoc config entries if set...
        if not kernel_spec["metadata"]["kernel_provisioner"].get("config"):
            kernel_spec["metadata"]["kernel_provisioner"]["config"] = {}
        self.add_optional_config_entries(kernel_spec["metadata"]["kernel_provisioner"]["config"])

        kernel_json_file = os.path.join(self.install_dir, KERNEL_JSON)
        self.log.debug(f"Finalizing kernel json file for kernel: '{self.display_name}'")
        with open(kernel_json_file, "w+") as f:
            json.dump(kernel_spec, f, indent=2)
            f.write("\n")

    def add_optional_config_entries(self, config_stanza: dict) -> None:
        """
        Adds optional configuration parameters to the 'config' stanza of 'kernel_provisioner'.

        This method is overridden by subclasses which should call super().add_optional_config_entries().
        """
        if (
            self.authorized_users
            and list(self.authorized_users) != self._authorized_users_default()
        ):
            config_stanza["authorized_users"] = list(self.authorized_users)
        if (
            self.unauthorized_users
            and list(self.unauthorized_users) != self._unauthorized_users_default()
        ):
            config_stanza["unauthorized_users"] = list(self.unauthorized_users)
        if self.port_range and self.port_range != self.port_range_default_value:
            config_stanza["port_range"] = self.port_range
        if self.launch_timeout:  # Always add launch_timeout if there's a value.
            config_stanza["launch_timeout"] = self.launch_timeout

    def get_substitutions(self, install_dir: os.path) -> dict:
        """
        Gather substitution strings to inject into the templated files.

        This method is overridden by subclasses which should first call super().get_substitutions().
        """
        substitutions = {}
        substitutions["display_name"] = self.display_name
        substitutions["install_dir"] = install_dir
        substitutions["language"] = LANGUAGE_SUBSTITUTIONS[self.language.lower()]
        substitutions["ipykernel_subclass_name"] = self.ipykernel_subclass_name
        return substitutions

    def _replace_existing(self) -> tuple[str, str]:
        temp_dir = ""
        destination_dir = self._get_destination_dir()
        kernel_json = os.path.join(destination_dir, "kernel.json")
        if os.path.exists(kernel_json):
            if not self.replace:
                err_msg = (
                    f"A kernel specification exists at '{destination_dir}' and '--replace' "
                    f"has not been specified.  Terminating."
                )
                raise SystemExit(err_msg)
            else:
                temp_dir = tempfile.mkdtemp()
                shutil.copytree(src=destination_dir, dst=temp_dir, dirs_exist_ok=True)
        return destination_dir, temp_dir

    def _get_destination_dir(self) -> str:
        """This method is essentially oa copy of `_get_destination_dir` from jupyter_client/kernelspec.py"""
        if self.user:
            return os.path.join(jupyter_data_dir(), "kernels", self.kernel_name)
        elif self.prefix:
            return os.path.join(
                os.path.abspath(self.prefix), "share", "jupyter", "kernels", self.kernel_name
            )
        else:
            return os.path.join(SYSTEM_JUPYTER_PATH[0], "kernels", self.kernel_name)


class BaseSpecSparkApp(BaseSpecApp):
    spark_home = Unicode(
        os.getenv("SPARK_HOME", "/opt/spark"),
        config=True,
        help="""Specify where the spark files can be found.""",
    )

    spark_init_mode = Unicode(
        DEFAULT_INIT_MODE,
        config=True,
        help=f"""Spark context initialization mode.  Must be one of {SPARK_INIT_MODES}.
    Default = {DEFAULT_INIT_MODE}.""",
    )

    @validate("spark_init_mode")
    def _spark_init_mode_validate(self, proposal: dict[str, str]) -> str:
        value = proposal["value"]
        try:
            assert value.lower() in SPARK_INIT_MODES
        except AssertionError as ae:
            err_msg = (
                f"Invalid Spark initialization mode value '{value}', not in {SPARK_INIT_MODES}"
            )
            raise TraitError(err_msg) from ae
        return value.lower()  # always use lowercase form

    extra_spark_opts = Unicode("", config=True, help="Specify additional Spark options.")

    spark = Bool(False, config=True, help="Install kernelspec for use with Spark.")

    aliases = {
        "spark-home": "BaseSpecSparkApp.spark_home",
        "spark-init-mode": "BaseSpecSparkApp.spark_init_mode",
        "extra-spark-opts": "BaseSpecSparkApp.extra_spark_opts",
    }
    aliases.update(BaseSpecApp.aliases)

    flags = {
        "spark": (
            {"BaseSpecSparkApp": {"spark": True}},
            spark.help,
        ),
    }
    flags.update(BaseSpecApp.flags)

    def get_substitutions(self, install_dir: os.path) -> dict:
        """
        Gather substitution strings to inject into the templated files.

        This method is overridden by subclasses which should first call super().get_substitutions().
        """
        substitutions = super().get_substitutions(install_dir)
        substitutions["spark_home"] = self.spark_home
        substitutions["extra_spark_opts"] = self.extra_spark_opts
        substitutions["spark_init_mode"] = self.spark_init_mode
        return substitutions



"""A spark operator provisioner."""
# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from .crd import CustomResourceProvisioner


class SparkOperatorProvisioner(CustomResourceProvisioner):
    """Spark operator provisioner."""

    # Identifies the kind of object being managed by this provisioner.
    # For these values we will prefer the values found in the 'kind' field
    # of the object's metadata.  This attribute is strictly used to provide
    # context to log messages.
    object_kind = "SparkApplication"

    def __init__(self, **kwargs):
        """Initialize the provisioner."""
        super().__init__(**kwargs)
        self.group = "sparkoperator.k8s.io"
        self.version = "v1beta2"
        self.plural = "sparkapplications"



#!/opt/conda/bin/python
# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Launch a custom operator resource."""
import argparse
import os
import sys

import urllib3
import yaml
from jinja2 import Environment, FileSystemLoader, select_autoescape
from kubernetes import client, config

urllib3.disable_warnings()


def generate_kernel_custom_resource_yaml(kernel_crd_template, keywords):
    j_env = Environment(
        loader=FileSystemLoader(os.path.dirname(__file__)),
        trim_blocks=True,
        lstrip_blocks=True,
        autoescape=select_autoescape(
            disabled_extensions=(
                "j2",
                "yaml",
            ),
            default_for_string=True,
            default=True,
        ),
    )
    k8s_yaml = j_env.get_template("/" + kernel_crd_template + ".yaml.j2").render(**keywords)
    return k8s_yaml


def extend_operator_env(op_def: dict, sub_spec: str) -> dict:
    """Extends the op_def.spec.sub_spec.env stanza with current environment."""
    env_stanza = op_def["spec"][sub_spec].get("env") or []

    # Walk current set of template env entries and replace those found in the current
    # env with their values (and record those items).   Then add all others from the env
    # that were not already.
    processed_entries: list[str] = []
    for item in env_stanza:
        item_name = item.get("name")
        if item_name in os.environ:
            item["value"] = os.environ[item_name]
            processed_entries.append(item_name)

    for name, value in os.environ.items():
        if name not in processed_entries:
            env_stanza.append({"name": name, "value": value})

    op_def["spec"][sub_spec]["env"] = env_stanza
    return op_def


def launch_custom_resource_kernel(
    kernel_id, port_range, response_addr, public_key, spark_context_init_mode, kernel_class_name
):
    config.load_incluster_config()

    keywords = {}

    keywords["port_range"] = port_range
    keywords["public_key"] = public_key
    keywords["response_address"] = response_addr
    keywords["kernel_id"] = kernel_id
    keywords["kernel_name"] = os.path.basename(os.path.dirname(os.path.dirname(__file__)))
    keywords["spark_context_initialization_mode"] = spark_context_init_mode
    keywords["kernel_class_name"] = kernel_class_name

    for name, value in os.environ.items():
        if name.startswith("KERNEL_"):
            keywords[name.lower()] = yaml.safe_load(value)

    kernel_crd_template = keywords["kernel_crd_group"] + "-" + keywords["kernel_crd_version"]
    custom_resource_yaml = generate_kernel_custom_resource_yaml(kernel_crd_template, keywords)

    kernel_namespace = keywords["kernel_namespace"]
    group = keywords["kernel_crd_group"]
    version = keywords["kernel_crd_version"]
    plural = keywords["kernel_crd_plural"]
    custom_resource_object = yaml.safe_load(custom_resource_yaml)

    if group == "sparkoperator.k8s.io":
        extend_operator_env(custom_resource_object, "driver")
        extend_operator_env(custom_resource_object, "executor")

    try:
        client.CustomObjectsApi().create_namespaced_custom_object(
            group, version, kernel_namespace, plural, custom_resource_object
        )
    except client.exceptions.ApiException as ex:
        if ex.status == 404:
            sys.exit(
                "\nERROR: The Kubernetes Operator for Apache Spark does not appear to be installed.  "
                "See 'https://github.com/GoogleCloudPlatform/spark-on-k8s-operator#installation' for "
                "instructions, then retry the operation.\n"
            )
        else:
            print("ERROR processing Kubernetes Operator CRD - kernel launch terminating!")
            print(custom_resource_yaml)
        raise ex


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--kernel-id",
        dest="kernel_id",
        nargs="?",
        help="Indicates the id associated with the launched kernel.",
    )
    parser.add_argument(
        "--port-range",
        dest="port_range",
        nargs="?",
        metavar="<lowerPort>..<upperPort>",
        help="Port range to impose for kernel ports",
    )
    parser.add_argument(
        "--response-address",
        dest="response_address",
        nargs="?",
        metavar="<ip>:<port>",
        help="Connection address (<ip>:<port>) for returning connection file",
    )
    parser.add_argument(
        "--public-key",
        dest="public_key",
        nargs="?",
        help="Public key used to encrypt connection information",
    )
    parser.add_argument(
        "--spark-context-initialization-mode",
        dest="spark_context_init_mode",
        nargs="?",
        help="Indicates whether or how a spark context should be created",
        default="none",
    )
    parser.add_argument(
        "--kernel-class-name",
        dest="kernel_class_name",
        nargs="?",
        help="Indicates the name of the kernel class to use.  Must be a subclass of 'ipykernel.kernelbase.Kernel'.",
    )
    arguments = vars(parser.parse_args())
    kernel_id = arguments["kernel_id"]
    port_range = arguments["port_range"]
    response_addr = arguments["response_address"]
    public_key = arguments["public_key"]
    spark_context_init_mode = arguments["spark_context_init_mode"]
    kernel_class_name = arguments["kernel_class_name"]

    launch_custom_resource_kernel(
        kernel_id, port_range, response_addr, public_key, spark_context_init_mode, kernel_class_name
    )



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Mixin for configuation options on RemoteProvisionerBase."""

import os

from tornado.log import LogFormatter
from traitlets import Integer, Set, Unicode, default
from traitlets.config import Configurable

# Commonly used envs
max_poll_attempts = int(os.getenv("GP_MAX_POLL_ATTEMPTS", "10"))
poll_interval = float(os.getenv("GP_POLL_INTERVAL", "0.5"))
socket_timeout = float(os.getenv("GP_SOCKET_TIMEOUT", "0.01"))
ssh_port = int(os.getenv("GP_SSH_PORT", "22"))


class RemoteProvisionerConfigMixin(Configurable):
    _log_formatter_cls = LogFormatter  # traitlet default is LevelFormatter

    @default("log_format")
    def _default_log_format(self):
        """override default log format to include milliseconds"""
        return (
            "%(color)s[%(levelname)1.1s %(asctime)s.%(msecs).03d %(name)s]%(end_color)s %(message)s"
        )

    # Authorized users
    authorized_users_env = "GP_AUTHORIZED_USERS"
    authorized_users = Set(
        config=True,
        help="""List of user names against which KERNEL_USERNAME will be compared.
Any match (case-sensitive) will allow the kernel's launch, otherwise an HTTP 403
(Forbidden) error will be raised.  The set of unauthorized users takes precedence.
This option should be used carefully as it can dramatically limit who can launch kernels.
To specify multiple names via the CLI, separate options must be provided for each entry.
(GP_AUTHORIZED_USERS env var - non-bracketed, just comma-separated)""",
    )

    @default("authorized_users")
    def _authorized_users_default(self):
        au_env = os.getenv(self.authorized_users_env)
        return set(au_env.split(",") if au_env is not None else [])

    # Unauthorized users
    unauthorized_users_env = "GP_UNAUTHORIZED_USERS"
    unauthorized_users_default_value = "root"
    unauthorized_users = Set(
        default_value={unauthorized_users_default_value},
        config=True,
        help="""List of user names against which KERNEL_USERNAME will be compared.
Any match (case-sensitive) will prevent the kernel's launch and result in an
HTTP 403 (Forbidden) error. To specify multiple names via the CLI, separate
options must be provided for each entry.
(GP_UNAUTHORIZED_USERS env var - non-bracketed, just comma-separated)""",
    )

    @default("unauthorized_users")
    def _unauthorized_users_default(self):
        return os.getenv(self.unauthorized_users_env, self.unauthorized_users_default_value).split(
            ","
        )

    # Port range
    port_range_env = "GP_PORT_RANGE"
    port_range_default_value = "0..0"
    port_range = Unicode(
        port_range_default_value,
        config=True,
        allow_none=True,
        help="""Specifies the lower and upper port numbers from which ports are created.
The bounded values are separated by '..' (e.g., 33245..34245 specifies a range of 1000 ports
to be randomly selected). A range of zero (e.g., 33245..33245 or 0..0) disables port-range
enforcement.  (GP_PORT_RANGE env var)""",
    )

    @default("port_range")
    def _port_range_default(self):
        return os.getenv(self.port_range_env, self.port_range_default_value)

    # # Conductor endpoint - TODO: Move to ConductorProvisioner when created
    # conductor_endpoint_env = 'GP_CONDUCTOR_ENDPOINT'
    # conductor_endpoint_default_value = None
    # conductor_endpoint = Unicode(conductor_endpoint_default_value,
    #                              allow_none=True,
    #                              config=True,
    #                              help="""The http url for accessing the Conductor REST API.
    #                              (GP_CONDUCTOR_ENDPOINT env var)""")
    #
    # @default('conductor_endpoint')
    # def conductor_endpoint_default(self):
    #     return os.getenv(self.conductor_endpoint_env, self.conductor_endpoint_default_value)

    launch_timeout_env = "GP_LAUNCH_TIMEOUT"
    launch_timeout_default_value = 30
    launch_timeout = Integer(
        launch_timeout_default_value,
        config=True,
        help="""Number of ports to try if the specified port is not available
(GP_LAUNCH_TIMEOUT env var)""",
    )

    @default("launch_timeout")
    def _launch_timeout_default(self):
        return int(
            os.getenv(
                self.launch_timeout_env,
                os.getenv("KERNEL_LAUNCH_TIMEOUT", self.launch_timeout_default_value),
            )
        )



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Response manager used by remote provisioners."""

import asyncio
import base64
import errno
import json
import os
import random
import re
from asyncio import Event
from socket import AF_INET, SO_REUSEADDR, SOCK_STREAM, SOL_SOCKET, socket, timeout
from typing import Any

from Cryptodome.Cipher import AES, PKCS1_v1_5
from Cryptodome.PublicKey import RSA
from Cryptodome.Util.Padding import unpad
from jupyter_client import localinterfaces
from tornado.ioloop import PeriodicCallback
from traitlets.config import SingletonConfigurable

from .config_mixin import poll_interval, socket_timeout

response_ip = os.getenv("GP_RESPONSE_IP", None)
desired_response_port = int(os.getenv("GP_RESPONSE_PORT", 8877))
response_port_retries = int(os.getenv("GP_RESPONSE_PORT_RETRIES", 10))
response_addr_any = bool(os.getenv("GP_RESPONSE_ADDR_ANY", "False").lower() == "true")

connection_interval = (
    poll_interval / 100.0
)  # already polling, so make connection timeout a fraction of outer poll


# Allow users to specify local ips (regular expressions can be used) that should not be included
# when determining the response address.  For example, on systems with many network interfaces,
# some may have their IPs appear the local interfaces list (e.g., docker's 172.17.0.* is an example)
# that should not be used.  This env can be used to indicate such IPs.
prohibited_local_ips = os.getenv("GP_PROHIBITED_LOCAL_IPS", "").split(",")


def _get_local_ip() -> str:
    """
    Honor the prohibited IPs, locating the first not in the list.
    """
    for ip in localinterfaces.public_ips():
        is_prohibited = False
        for prohibited_ip in prohibited_local_ips:  # exhaust prohibited list, applying regexs
            if re.match(prohibited_ip, ip):
                is_prohibited = True
                break
        if not is_prohibited:
            return ip
    return localinterfaces.public_ips()[0]  # all were prohibited, so go with the first


local_ip = _get_local_ip()

random.seed()


class Response(Event):
    """Combines the event behavior with the kernel launch response."""

    _response = None

    @property
    def response(self):
        return self._response

    @response.setter
    def response(self, value):
        """Set the response.  NOTE: this marks the event as set."""
        self._response = value
        self.set()


class ResponseManager(SingletonConfigurable):
    """
    Singleton that manages the responses from each kernel launcher at startup.

    This singleton does the following:
    1. Acquires a public and private RSA key pair at first use to encrypt and decrypt the
    received responses.  The public key is sent to the launcher during startup
    and is used by the launcher to encrypt the AES key the launcher uses to encrypt
    the connection information, while the private key remains in the server and is
    used to decrypt the AES key from the response - which it then uses to decrypt
    the connection information.
    2. Creates a single socket based on the configuration settings that is listened on
    via a periodic callback.
    3. On receipt, it decrypts the response (key then connection info) and posts the
    response payload to a map identified by the kernel_id embedded in the response.
    4. Provides a wait mechanism for callers to poll to get their connection info
    based on their registration (of kernel_id).
    """

    KEY_SIZE = 1024  # Can be small since its' only used to {en,de}crypt the AES key.
    _instance = None

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        self._response_ip = None
        self._response_port = None
        self._response_socket = None
        self._connection_processor = None

        # Create encryption keys...
        self._private_key = RSA.generate(ResponseManager.KEY_SIZE)
        self._public_key = self._private_key.publickey()
        self._public_pem = self._public_key.export_key("PEM")

        # Event facility...
        self._response_registry = {}

        # Start the response manager (create socket, periodic callback, etc.) ...
        self._start_response_manager()

    @property
    def public_key(self) -> str:
        """Provides the string-form of public key PEM with header/footer/newlines stripped."""
        return (
            self._public_pem.decode()
            .replace("-----BEGIN PUBLIC KEY-----", "")
            .replace("-----END PUBLIC KEY-----", "")
            .replace("\n", "")
        )

    @property
    def response_address(self) -> str:
        return f"{self._response_ip}:{self._response_port}"

    def register_event(self, kernel_id: str) -> None:
        """Register kernel_id so its connection information can be processed."""
        self._response_registry[kernel_id] = Response()

    async def get_connection_info(self, kernel_id: str) -> dict:
        """Performs a timeout wait on the event, returning the connection information on completion."""
        await asyncio.wait_for(self._response_registry[kernel_id].wait(), connection_interval)
        return self._response_registry.pop(kernel_id).response

    def _prepare_response_socket(self):
        """Prepares the response socket on which connection info arrives from remote kernel launcher."""
        s = socket(AF_INET, SOCK_STREAM)
        s.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)

        # If response_addr_any is enabled (default disabled), we will permit the server to listen
        # on all addresses, else we will honor a configured response IP (via env) over the local IP
        # (which is the default).
        # Multiple IP bindings should be configured for containerized configurations (k8s) that need to
        # launch kernels into external YARN clusters.
        bind_ip = local_ip if response_ip is None else response_ip
        bind_ip = bind_ip if response_addr_any is False else ""

        response_port = desired_response_port
        for port in ResponseManager._random_ports(response_port, response_port_retries + 1):
            try:
                s.bind((bind_ip, port))
            except OSError as e:
                if e.errno == errno.EADDRINUSE:
                    self.log.info(f"Response port {port} is already in use, trying another port...")
                    continue
                elif e.errno in (
                    errno.EACCES,
                    getattr(errno, "WSAEACCES", errno.EACCES),
                ):
                    self.log.warning(
                        f"Permission to bind to response port {port} denied - continuing..."
                    )
                    continue
                else:
                    err_msg = f"Failed to bind to port '{port}' for response address due to: '{e}'"
                    raise RuntimeError(err_msg) from e
            else:
                response_port = port
                break
        else:
            msg = f"No available response port could be found after {response_port_retries + 1} attempts"
            self.log.critical(msg)
            raise RuntimeError(msg)

        self.log.info(
            f"ResponseManager is bound to port {response_port} for remote kernel connection information."
        )
        s.listen(128)
        s.settimeout(socket_timeout)
        self._response_socket = s
        self._response_port = response_port
        self._response_ip = local_ip if response_ip is None else response_ip

    def _start_response_manager(self) -> None:
        """If not already started, creates and starts the periodic callback to process connections."""
        if self._response_socket is None:
            self._prepare_response_socket()

        if self._connection_processor is None:
            self._connection_processor = PeriodicCallback(self._process_connections, 0.1, 0.1)
            self._connection_processor.start()

    def stop_response_manager(self) -> None:
        """Stops the connection processor."""
        if self._connection_processor is not None:
            self._connection_processor.stop()
            self._connection_processor = None

        if self._response_socket is not None:
            self._response_socket = None

    async def _process_connections(self) -> None:
        """Checks the socket for data, if found, decrypts the payload and posts to 'wait map'."""
        loop = asyncio.get_event_loop()
        data = ""
        conn = None
        try:
            conn, addr = await loop.sock_accept(self._response_socket)
            while True:
                buffer = await loop.sock_recv(conn, 1024)
                if not buffer:  # send is complete, process payload
                    self.log.debug(f"Received payload '{data}'")
                    payload = self._decode_payload(data)
                    self.log.debug(f"Decrypted payload '{payload}'")
                    self._post_connection(payload)
                    break
                data = data + buffer.decode(
                    encoding="utf-8"
                )  # append what we received until we get no more...
        except timeout:
            pass
        except Exception as ex:
            self.log.error(f"Failure occurred processing connection: {ex}")
        if conn:
            conn.close()

    def _decode_payload(self, data) -> dict:
        """
        Decodes the payload.

        Decodes the payload, identifying the payload's version and returns a dictionary
        representing the kernel's connection information.

        Version "0" payloads do not specify a kernel-id within the payload, nor do they
        include a 'key', 'version' or 'conn_info' fields.  They are purely an AES encrypted
        form of the base64-encoded JSON connection information, and encrypted using the
        kernel-id as a key.  Since no kernel-id is in the payload, we will capture the keys
        of registered kernel-ids and attempt to decrypt the payload until we find the
        appropriate registrant.

        Version "1+" payloads are a base64-encoded JSON string consisting of a 'version', 'key'
        and 'conn_info' fields.  The 'key' field will be decrpyted using the private key to
        reveal the AES key, which is then used to decrypt the `conn_info` field.

        Once decryption has taken place, the connection information string is loaded into a
        dictionary and returned.
        """

        payload_str = base64.b64decode(data)
        try:
            payload = json.loads(payload_str)
            # Get the version
            version = payload.get("version")
            if version is None:
                err_msg = "Payload received from kernel does not include a version indicator!"
                raise ValueError(err_msg)
            self.log.debug(f"Version {version} payload received.")

            if version == 1:
                # Decrypt the AES key using the RSA private key
                encrypted_aes_key = base64.b64decode(payload["key"].encode())
                cipher = PKCS1_v1_5.new(self._private_key)
                aes_key = cipher.decrypt(encrypted_aes_key, b"\x42")
                # Per docs, don't convey that decryption returned sentinel.  So just let
                # things fail "naturally".
                # Decrypt and unpad the connection information using the just-decrypted AES key
                cipher = AES.new(aes_key, AES.MODE_ECB)
                encrypted_connection_info = base64.b64decode(payload["conn_info"].encode())
                connection_info_str = unpad(cipher.decrypt(encrypted_connection_info), 16).decode()
            else:
                err_msg = f"Unexpected version indicator received: {version}!"
                raise ValueError(err_msg)
        except Exception as ex:
            # Could be version "0", walk the registrant kernel-ids and attempt to decrypt using each as a key.
            # If none are found, re-raise the triggering exception.
            self.log.debug(f"decode_payload exception - {ex.__class__.__name__}: {ex}")
            connection_info_str = None
            for kernel_id in self._response_registry.keys():
                aes_key = kernel_id[0:16]
                try:
                    cipher = AES.new(aes_key.encode("utf-8"), AES.MODE_ECB)
                    decrypted_payload = cipher.decrypt(payload_str)
                    # Version "0" responses use custom padding, so remove that here.
                    connection_info_str = "".join(
                        [decrypted_payload.decode("utf-8").rsplit("}", 1)[0], "}"]
                    )
                    # Try to load as JSON
                    new_connection_info = json.loads(connection_info_str)
                    # Add kernel_id into dict, then dump back to string so this can be processed as valid response
                    new_connection_info["kernel_id"] = kernel_id
                    connection_info_str = json.dumps(new_connection_info)
                    self.log.warning(
                        f"WARNING!!!! Legacy kernel response received for kernel_id '{kernel_id}'! "
                        "Update kernel launchers to current version!"
                    )
                    break  # If we're here, we made it!
                except Exception as ex2:
                    # Any exception fails this experiment and we continue
                    self.log.debug(
                        f"Received the following exception detecting legacy kernel "
                        f"response - {ex2.__class__.__name__}: {ex2}"
                    )
                    connection_info_str = None

            if connection_info_str is None:
                raise ex

        # and convert to usable dictionary
        connection_info = json.loads(connection_info_str)
        if "key" in connection_info:  # Convert key to bytes
            connection_info["key"] = connection_info["key"].encode()
        return connection_info

    def _post_connection(self, connection_info: dict) -> None:
        """Posts connection information into "wait map" based on kernel_id value."""
        kernel_id = connection_info.get("kernel_id")
        if kernel_id is None:
            self.log.error("No kernel id found in response!  Kernel launch will fail.")
            return
        if kernel_id not in self._response_registry:
            self.log.error(
                f"Kernel id '{kernel_id}' has not been registered and will not be processed!"
            )
            return

        self.log.debug(f"Connection info received for kernel '{kernel_id}': {connection_info}")
        self._response_registry[kernel_id].response = connection_info

    @staticmethod
    def _random_ports(port: int, n):
        """Generate a list of n random ports near the given port.

        The first 5 ports will be sequential, and the remaining n-5 will be
        randomly selected in the range [port-2*n, port+2*n].
        """
        for i in range(min(5, n)):
            yield port + i
        for _ in range(n - 5):
            yield max(1, port + random.randint(-2 * n, 2 * n))



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
import os

from jupyter_core.application import JupyterApp
from overrides import overrides
from traitlets import List, Unicode, default

from .._version import __version__
from .base_app import DEFAULT_LANGUAGE, PYTHON, SCALA, BaseSpecSparkApp, R

DEFAULT_KERNEL_NAMES = {PYTHON: "ssh_python", SCALA: "ssh_scala", R: "ssh_r"}
DEFAULT_DISPLAY_NAMES = {PYTHON: "Python SSH", SCALA: "Scala SSH", R: "R SSH"}
SPARK_SUFFIX = "_spark"
SPARK_DISPLAY_NAME_SUFFIX = " (with Spark)"


class SshSpecInstaller(BaseSpecSparkApp):
    """CLI for extension management."""

    name = "jupyter-ssh-spec"
    description = (
        "Creates a Jupyter kernel specification for use within a cluster of hosts via SSH."
    )
    # Note that the left justification of the second example is necessary to ensure proper
    # alignment with the first example during --help output.
    examples = """
    jupyter-ssh-spec install --language R --spark --spark-master spark://192.168.2.5:7077 --spark-home=/usr/local/spark

jupyter-ssh-spec install --kernel-name ssh_python --remote-hosts=192.168.2.4 --remote-hosts=192.168.2.5

jupyter-ssh-spec install --language Scala --spark --spark_init_mode 'eager'
    """

    @default("kernel_name")
    def _kernel_name_default(self) -> str:
        return DEFAULT_KERNEL_NAMES[DEFAULT_LANGUAGE]

    @default("display_name")
    def _display_name_default(self) -> str:
        return DEFAULT_DISPLAY_NAMES[DEFAULT_LANGUAGE]

    remote_hosts_env = "GP_REMOTE_HOSTS"
    remote_hosts_default_value = "localhost"
    remote_hosts = List(
        default_value=[remote_hosts_default_value],
        config=True,
        help="""List of host names on which this kernel can be launched.  Multiple entries must
each be specified via separate options: --remote-hosts host1 --remote-hosts host2""",
    )

    @default("remote_hosts")
    def _remote_hosts_default(self):
        return os.getenv(self.remote_hosts_env, self.remote_hosts_default_value).split(",")

    spark_master_default_value = "yarn"
    spark_master = Unicode(
        default_value=spark_master_default_value,
        config=True,
        help="Specify the Spark Master URL (e.g., 'spark://HOST:PORT' or 'yarn'.",
    )
    # Flags

    aliases = {
        "remote-hosts": "SshSpecInstaller.remote_hosts",
        "spark-master": "SshSpecInstaller.spark_master",
    }
    aliases.update(BaseSpecSparkApp.aliases)

    flags = {}
    flags.update(BaseSpecSparkApp.flags)

    @overrides
    def validate_parameters(self):
        super().validate_parameters()

        self.language = self.language.lower()
        self.launcher_dir_name = self.language
        self.resource_dir_name = self.language

        if self.spark is True:
            # if kernel and display names are still defaulted, silently convert to lang default and append spark suffix
            if self.kernel_name == DEFAULT_KERNEL_NAMES[DEFAULT_LANGUAGE]:
                self.kernel_name = DEFAULT_KERNEL_NAMES[self.language] + SPARK_SUFFIX
            if self.display_name == DEFAULT_DISPLAY_NAMES[DEFAULT_LANGUAGE]:
                self.display_name = DEFAULT_DISPLAY_NAMES[self.language] + SPARK_DISPLAY_NAME_SUFFIX

            self.kernel_spec_dir_name = DEFAULT_KERNEL_NAMES[self.language] + SPARK_SUFFIX
        else:
            # if kernel and display names are still defaulted, silently change to language defaults
            if self.kernel_name == DEFAULT_KERNEL_NAMES[DEFAULT_LANGUAGE]:
                self.kernel_name = DEFAULT_KERNEL_NAMES[self.language]
            if self.display_name == DEFAULT_DISPLAY_NAMES[DEFAULT_LANGUAGE]:
                self.display_name = DEFAULT_DISPLAY_NAMES[self.language]
            self.kernel_spec_dir_name = DEFAULT_KERNEL_NAMES[self.language]

            self.spark_init_mode = "none"
            if len(self.extra_spark_opts) > 0:
                self.log.warning(
                    "--extra_spark_opts will be ignored since --spark has not been specified."
                )
                self.extra_spark_opts = ""

        # sanitize kernel_name
        self.kernel_name = self.kernel_name.replace(" ", "_")

    @overrides
    def add_optional_config_entries(self, config_stanza: dict) -> None:
        super().add_optional_config_entries(config_stanza)
        if self.remote_hosts and list(self.remote_hosts) != self._remote_hosts_default():
            config_stanza["remote_hosts"] = list(self.remote_hosts)

    @overrides
    def get_substitutions(self, install_dir) -> dict:
        substitutions = super().get_substitutions(install_dir)
        substitutions["spark_master"] = self.spark_master
        return substitutions


class SshProvisionerApp(JupyterApp):
    """Application responsible for driving the creation of Ssh-based kernel specifications."""

    version = __version__
    name = "jupyter ssh-spec"
    description = """Application used to create kernel specifications for use on clusters via SSH
    and the DistributedProvisioner kernel provisioner."""
    subcommands = dict(
        {
            "install": (SshSpecInstaller, SshSpecInstaller.description.splitlines()[0]),
        }
    )
    aliases = {}
    flags = {}

    def start(self):
        super().start()

        if self.subapp is None:
            print(f"No subcommand specified. Must specify one of: {list(self.subcommands)}")
            print()
            self.print_description()
            self.print_subcommands()
            self.exit(1)


if __name__ == "__main__":
    SshProvisionerApp.launch_instance()



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from dataclasses import dataclass

from kubernetes.client.rest import ApiException


class K8sResource:
    """Track the current state of the resource"""

    def __init__(self, env: dict):
        self.env = env
        self.kernel_id: str = env.get("KERNEL_ID")
        self.kernel_username: str = env.get("KERNEL_USERNAME")
        self.pod_name = f"{self.kernel_username}-{self.kernel_id}"
        self.namespace: str = env.get("KERNEL_NAMESPACE")
        self.status: str = "Pending"
        self.query_counter: int = 1


k8s_resources: dict = {}


@dataclass
class MockPodStatus:
    pod_ip: str
    host_ip: str
    phase: str


@dataclass
class MockPodMetadata:
    name: str


@dataclass
class MockPodInfo:
    status: MockPodStatus
    metadata: MockPodMetadata


class MockResponse:
    def __init__(self, pod_info: MockPodInfo):
        self.items: list[MockPodInfo] = []
        self.items.append(pod_info)


class MockCoreV1Api:
    def list_namespaced_pod(self, namespace, **kwargs):
        kernel_id: str = ""
        label_selector = kwargs.get("label_selector", "")
        selector_entries = label_selector.split(",")
        for entry in selector_entries:
            if entry.startswith("kernel_id="):
                kernel_id = entry.split("=")[1]
                break
        if kernel_id in k8s_resources:
            resource = k8s_resources.get(kernel_id)
            if resource.query_counter >= 3:  # time to return
                resource.status = "Running"
                pod_info = MockPodInfo(
                    status=MockPodStatus("127.0.0.1", "127.0.0.1", resource.status),
                    metadata=MockPodMetadata(name=resource.pod_name),
                )
                response = MockResponse(pod_info=pod_info)
                return response
            else:
                resource.status = "Pending"
                resource.query_counter += 1

        return None

    def delete_namespaced_pod(self, name, namespace, **kwargs):
        pod_info = None
        delete_resource = None
        for kid, resource in k8s_resources.items():
            if resource.pod_name == name:
                resource.status = "Terminating"
                delete_resource = kid
                pod_info = MockPodInfo(
                    status=MockPodStatus("127.0.0.1", "127.0.0.1", resource.status),
                    metadata=MockPodMetadata(name=resource.pod_name),
                )
                break

        if pod_info:
            k8s_resources.pop(delete_resource)
            return pod_info

        raise ApiException(status=404, reason="Could not find resource with pod-name: '{name}'!")

    def delete_namespace(self, name, body):
        # TODO - add impl when adding namespace lifecycle testing
        pass

    def create_namespace(self, body):
        # TODO - add impl when adding namespace lifecycle testing
        pass


class MockRbacAuthorizationV1Api:
    def create_namespaced_role_binding(self, namespace, body):
        # TODO - add impl when adding namespace lifecycle testing
        pass


class MockK8sClient:
    def __init__(self, **kwargs):
        self.args = kwargs

    @classmethod
    def CoreV1Api(cls):
        return MockCoreV1Api()

    @classmethod
    def RbacAuthorizationV1Api(cls):
        return MockRbacAuthorizationV1Api()

    @classmethod
    def V1DeleteOptions(cls, grace_period_seconds=0, propagation_policy="Background") -> dict:
        return {"grace_period_seconds": 0, "propagation_policy": "Background"}

    @classmethod
    def V1ObjectMeta(cls, name, labels) -> dict:
        pass

    @classmethod
    def V1Namespace(cls, metadata) -> dict:
        pass

    @classmethod
    def V1RoleRef(cls, api_group, name) -> dict:
        pass

    @classmethod
    def V1Subject(cls, api_group, kind, name, namespace) -> dict:
        pass

    @classmethod
    def V1RoleBinding(cls, kind, metadata, role_ref, subjects) -> dict:
        pass



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Code related to managing kernels running in YARN clusters."""
import asyncio
import getpass
import json
import os
import signal
import subprocess
import warnings
from socket import gethostbyname, gethostname
from typing import Any, Dict, Optional
from typing import List as tyList

import paramiko
from jupyter_client import KernelConnectionInfo, launch_kernel
from overrides import overrides
from traitlets import List, TraitError, Unicode, default, validate

from .config_mixin import max_poll_attempts, poll_interval, ssh_port
from .remote_provisioner import RemoteProvisionerBase

# would prefer /var/log, but its only writable by root
kernel_log_dir = os.getenv("GP_KERNEL_LOG_DIR", "/tmp")  # noqa: S108


class TrackKernelOnHost:
    """
    Class used to track the number of active kernels on the set of hosts
    so that the least-utilized host can be used for the next distributed
    request.
    """

    _host_kernels = {}
    _kernel_host_mapping = {}

    def add_kernel_id(self, host: str, kernel_id: str) -> None:
        self._kernel_host_mapping[kernel_id] = host
        self.increment(host)

    def delete_kernel_id(self, kernel_id: str) -> None:
        host = self._kernel_host_mapping.get(kernel_id)
        if host:
            self.decrement(host)
            del self._kernel_host_mapping[kernel_id]

    def min_or_remote_host(self, remote_host: Optional[str] = None) -> str:
        if remote_host:
            return remote_host
        return min(self._host_kernels, key=lambda k: self._host_kernels[k])

    def increment(self, host: str) -> None:
        val = int(self._host_kernels.get(host, 0))
        self._host_kernels[host] = val + 1

    def decrement(self, host: str) -> None:
        val = int(self._host_kernels.get(host, 0))
        self._host_kernels[host] = val - 1

    def init_host_kernels(self, hosts) -> None:
        if len(self._host_kernels) == 0:
            self._host_kernels.update({key: 0 for key in hosts})


class DistributedProvisioner(RemoteProvisionerBase):
    """
    Kernel lifecycle management for clusters via ssh and a set of hosts.
    """

    host_index = 0
    kernel_on_host = TrackKernelOnHost()

    remote_hosts_env = "GP_REMOTE_HOSTS"
    remote_hosts_default_value = "localhost"
    remote_hosts = List(
        default_value=[remote_hosts_default_value],
        config=True,
        help="""List of host names on which this kernel can be launched.  Multiple entries must
each be specified via separate options: --remote-hosts host1 --remote-hosts host2""",
    )

    @default("remote_hosts")
    def _remote_hosts_default(self):
        return os.getenv(self.remote_hosts_env, self.remote_hosts_default_value).split(",")

    load_balancing_algorithm_env = "GP_LOAD_BALANCING_ALGORITHM"
    load_balancing_algorithm_default_value = "round-robin"
    load_balancing_algorithm = Unicode(
        load_balancing_algorithm_default_value,
        config=True,
        help="""Specifies which load balancing algorithm DistributedProvisioner should use.
Must be one of "round-robin" or "least-connection".  (GP_LOAD_BALANCING_ALGORITHM env var)""",
    )

    @default("load_balancing_algorithm")
    def _load_balancing_algorithm_default(self) -> str:
        return os.getenv(
            self.load_balancing_algorithm_env,
            self.load_balancing_algorithm_default_value,
        )

    @validate("load_balancing_algorithm")
    def _validate_load_balancing_algorithm(self, proposal: Dict[str, str]) -> str:
        value = proposal["value"]
        try:
            assert value in ["round-robin", "least-connection"]
        except ValueError as ve:
            err_msg = f"Invalid load_balancing_algorithm value {value}, not in [round-robin,least-connection]"
            raise TraitError(err_msg) from ve
        return value

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.kernel_log = None
        self.local_stdout = None
        self.least_connection = self.load_balancing_algorithm == "least-connection"
        _remote_user = os.getenv("GP_REMOTE_USER")
        self.remote_pwd = os.getenv("GP_REMOTE_PWD")
        self.use_gss = os.getenv("GP_REMOTE_GSS_SSH", "False").lower() == "true"
        if self.use_gss:
            if self.remote_pwd or _remote_user:
                warnings.warn(
                    "Both `GP_REMOTE_GSS_SSH` and one of `GP_REMOTE_PWD` or `GP_REMOTE_USER` is set. "
                    "Those options are mutually exclusive, you configuration may be incorrect. "
                    "GP_REMOTE_GSS_SSH will take priority.",
                    stacklevel=2,
                )
            self.remote_user = None
        else:
            self.remote_user = _remote_user if _remote_user else getpass.getuser()

    @property
    @overrides
    def has_process(self) -> bool:
        return self.local_proc is not None or (self.ip is not None and self.pid > 0)

    @overrides
    async def launch_kernel(self, cmd: tyList[str], **kwargs: Any) -> KernelConnectionInfo:
        """
        Launches a kernel process on a selected host.

        NOTE: This overrides the superclass `launch_kernel` method entirely.
        """
        self.kernel_log = None
        env_dict = kwargs.get("env", {})
        self.assigned_host = self._determine_next_host(env_dict)
        self.ip = gethostbyname(self.assigned_host)  # convert to ip if host is provided
        self.assigned_ip = self.ip

        launch_kwargs = RemoteProvisionerBase._scrub_kwargs(kwargs)
        try:
            result_pid = self._launch_remote_process(cmd, **launch_kwargs)
            self.pid = int(result_pid)
        except Exception as e:
            error_message = (
                f"Failure occurred starting kernel on '{self.ip}'.  Returned result: {e}"
            )
            self.log_and_raise(RuntimeError(error_message))

        self.log_kernel_launch(cmd)

        await self.confirm_remote_startup()

        return self.connection_info

    @overrides
    async def poll(self) -> Optional[int]:
        signal_delivered = await self._send_signal_via_listener(0)
        if signal_delivered:  # kernel process still alive, return None
            return None
        return 0

    @overrides
    async def kill(self, restart=False) -> None:
        # If we have a local process, use its method, else signal soft kill first before hard kill.
        res = await self.poll()
        if res is not None:  # Already terminated
            self.log.debug("Distributed: kill: already terminated.")
            return None
        await self.terminate()  # Send -15 signal first
        i = 1
        while await self.poll() is None and i <= max_poll_attempts:
            await asyncio.sleep(poll_interval)
            i = i + 1
        if i > max_poll_attempts:  # Send -9 signal if process is still alive
            await self.send_signal(signal.SIGKILL)

    @overrides
    async def terminate(self, restart=False) -> None:
        await self.send_signal(signal.SIGTERM)

    @overrides
    async def cleanup(self, restart=False) -> None:
        self._unregister_assigned_host()
        if self.local_stdout:
            self.local_stdout.close()
            self.local_stdout = None
        await super().cleanup()

    @overrides
    async def confirm_remote_startup(self):
        self.start_time = RemoteProvisionerBase.get_current_time()
        i = 0
        ready_to_connect = False  # we're ready to connect when we have a connection file to use
        while not ready_to_connect:
            i += 1
            await self.handle_launch_timeout()

            self.log.debug(
                "{}: Waiting to connect.  Host: '{}', KernelID: '{}'".format(
                    i, self.assigned_host, self.kernel_id
                )
            )

            if self.assigned_host != "":
                ready_to_connect = await self.receive_connection_info()

    @overrides
    def log_kernel_launch(self, cmd: tyList[str]) -> None:
        self.log.info(
            f"{self.__class__.__name__}: kernel launched.  Host: '{self.assigned_host}', "
            f"pid: {self.pid}, Kernel ID: {self.kernel_id}, "
            f"Log file: {self.assigned_host}:{self.kernel_log}, cmd: '{cmd}'."
        )

    def _launch_remote_process(self, cmd: tyList[str], **kwargs: Any):
        """
        Launch the kernel as indicated by the argv stanza in the kernelspec.  Note that this method
        will bypass use of ssh if the remote host is also the local machine.
        """

        cmd = self._build_startup_command(cmd, **kwargs)
        self.log.debug(f"Invoking cmd: '{cmd}' on host: {self.assigned_host}")
        result_pid = "bad_pid"  # purposely initialize to bad int value

        if RemoteProvisionerBase.ip_is_local(self.ip):
            # launch the local command with redirection in place
            self.local_stdout = open(self.kernel_log, mode="a")
            self.local_proc = launch_kernel(
                cmd, stdout=self.local_stdout, stderr=subprocess.STDOUT, **kwargs
            )
            result_pid = str(self.local_proc.pid)
        else:
            # launch remote command via ssh
            result = self._rsh(self.ip, "".join(cmd))
            for line in result:
                result_pid = line.strip()

        return result_pid

    def _build_startup_command(self, cmd: tyList[str], **kwargs: Any) -> tyList[str]:
        """
        Builds the command to invoke by concatenating envs from kernelspec followed by the kernel argvs.

        We also force nohup, redirection to a file and place in background, then follow with an echo
        for the background pid.

        Note: We optimize for the local case and just return the existing command.
        """

        # Optimized case needs to also redirect the kernel output, so unconditionally compose kernel_log
        env_dict = kwargs["env"]
        kid = env_dict.get("KERNEL_ID")
        self.kernel_log = os.path.join(kernel_log_dir, f"kernel-{kid}.log")

        if RemoteProvisionerBase.ip_is_local(self.ip):  # We're local so just use what we're given
            startup_cmd = cmd
        else:  # Add additional envs, including those in kernelspec
            startup_cmd = ""
            if kid:
                startup_cmd += f'export KERNEL_ID="{kid}";'

            kernel_user = env_dict.get("KERNEL_USERNAME")
            if kernel_user:
                startup_cmd += f'export KERNEL_USERNAME="{kernel_user}";'

            impersonation = env_dict.get("GP_IMPERSONATION_ENABLED")
            if impersonation:
                startup_cmd += f'export GP_IMPERSONATION_ENABLED="{impersonation}";'

            for key, value in self.kernel_spec.env.items():
                startup_cmd += "export {}={};".format(key, json.dumps(value).replace("'", "''"))

            startup_cmd += "nohup"
            for arg in cmd:
                startup_cmd += f" {arg}"

            startup_cmd += f" >> {self.kernel_log} 2>&1 & echo $!"  # return the process id

        return startup_cmd

    def _determine_next_host(self, env_dict: dict) -> str:
        """Simple round-robin index into list of hosts."""
        remote_host = env_dict.get("KERNEL_REMOTE_HOST")
        if self.least_connection:
            next_host = DistributedProvisioner.kernel_on_host.min_or_remote_host(remote_host)
            DistributedProvisioner.kernel_on_host.add_kernel_id(next_host, self.kernel_id)
        else:
            next_host = (
                remote_host
                if remote_host
                else self.remote_hosts[
                    DistributedProvisioner.host_index % self.remote_hosts.__len__()
                ]
            )
            DistributedProvisioner.host_index += 1

        return next_host

    def _unregister_assigned_host(self) -> None:
        if self.least_connection:
            DistributedProvisioner.kernel_on_host.delete_kernel_id(self.kernel_id)

    def _get_ssh_client(self, host) -> paramiko.SSHClient:
        """
        Create a SSH Client based on host, username and password if provided.
        If there is any AuthenticationException/SSHException, raise HTTP Error 403 as permission denied.

        :param host:
        :return: ssh client instance
        """
        ssh = None

        try:
            ssh = paramiko.SSHClient()
            ssh.load_system_host_keys()
            host_ip = gethostbyname(host)
            if self.use_gss:
                self.log.debug("Connecting to remote host via GSS.")
                ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                ssh.connect(host_ip, port=ssh_port, gss_auth=True)
            else:
                ssh.set_missing_host_key_policy(paramiko.RejectPolicy())
                if self.remote_pwd:
                    self.log.debug("Connecting to remote host with username and password.")
                    ssh.connect(
                        host_ip,
                        port=ssh_port,
                        username=self.remote_user,
                        password=self.remote_pwd,
                    )
                else:
                    self.log.debug("Connecting to remote host with ssh key.")
                    ssh.connect(host_ip, port=ssh_port, username=self.remote_user)
        except Exception as e:
            current_host = gethostbyname(gethostname())
            error_message = (
                f"Exception '{type(e).__name__}' occurred when creating a SSHClient at {current_host} connecting "
                f"to '{host}:{ssh_port}' with user '{self.remote_user}', message='{e}'."
            )
            if e is paramiko.SSHException or paramiko.AuthenticationException:
                error_message_prefix = "Failed to authenticate SSHClient with password"
                error_message = error_message_prefix + (
                    " provided" if self.remote_pwd else "-less SSH"
                )
                error_message = error_message + f"and GP_REMOTE_GSS_SSH={self.use_gss}"

            self.log_and_raise(RuntimeError(error_message))

        return ssh

    def _rsh(self, host, command):
        """
        Executes a command on a remote host using ssh.

        Parameters
        ----------
        host : str
            The host on which the command is executed.
        command : str
            The command to execute.

        Returns
        -------
        lines : List
            The command's output.  If stdout is zero length, the stderr output is returned.
        """
        ssh = self._get_ssh_client(host)
        try:
            stdin, stdout, stderr = ssh.exec_command(command, timeout=30)
            lines = stdout.readlines()
            if len(lines) == 0:  # if nothing in stdout, return stderr
                lines = stderr.readlines()
        except Exception as e:
            # Let caller decide if exception should be logged
            raise e

        finally:
            if ssh:
                ssh.close()

        return lines



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Code related to managing kernels running in Kubernetes clusters."""

import logging
import os
import re
from typing import Any, Dict, List, Optional, Set

import urllib3
from overrides import overrides

try:
    from jinja2 import Environment  # noqa - used by launcher so check presence here
    from kubernetes import client, config
    from kubernetes.client.rest import ApiException
except ImportError:
    logging.warning(
        "At least one of the extra packages 'kubernetes' or 'jinja2' are not installed in "
        "this environment and are required.  Ensure that gateway_provisioners is installed "
        "by specifying the extra 'k8s' (e.g., pip install 'gateway_provisioners[k8s]')."
    )
    raise


from .container import ContainerProvisionerBase

urllib3.disable_warnings()

# Default logging level of kubernetes produces too much noise - raise to warning only.
logging.getLogger("kubernetes").setLevel(os.environ.get("GP_KUBERNETES_LOG_LEVEL", logging.WARNING))

k8s_provisioner_namespace = os.environ.get("GP_NAMESPACE", "default")
default_kernel_service_account_name = os.environ.get(
    "GP_DEFAULT_KERNEL_SERVICE_ACCOUNT_NAME", "default"
)
kernel_cluster_role = os.environ.get("GP_KERNEL_CLUSTER_ROLE", "cluster-admin")

# Since provisioners are a single-user scenario (not going through EG), use a shared namespace by default.
shared_namespace = bool(os.environ.get("GP_SHARED_NAMESPACE", "True").lower() == "true")
kpt_dir = os.environ.get("GP_POD_TEMPLATE_DIR", "/tmp")  # noqa: S108

app_name = os.environ.get("GP_APP_NAME", "gateway-provisioners")

if (
    "SPHINX_BUILD_IN_PROGRESS" not in os.environ
    and "PYTEST_CURRENT_TEST" not in os.environ
    and "PYTEST_RUN_CONFIG" not in os.environ
):
    if bool(os.environ.get("GP_USE_INCLUSTER_CONFIG", "True").lower() == "true"):
        config.load_incluster_config()
    else:
        config.load_kube_config()


class KubernetesProvisioner(ContainerProvisionerBase):
    """Kernel lifecycle management for Kubernetes kernels."""

    # Identifies the kind of object being managed by this provisioner.
    # For these values we will prefer the values found in the 'kind' field
    # of the object's metadata.  This attribute is strictly used to provide
    # context to log messages.
    object_kind = "Pod"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.kernel_pod_name = None
        self.kernel_namespace = None
        self.delete_kernel_namespace = False
        # Track if we're restarting the pod from this instance.  This will be
        # set to true when restarting kernels during the original pod's termination
        # and if the kernel resides in its own namespace.
        self.restarting = False

    @overrides
    async def pre_launch(self, **kwargs: Any) -> Dict[str, Any]:
        # Set env before superclass call, so we can see these in the debug output

        # Kubernetes relies on internal env variables to determine its configuration.  When
        # running within a K8s cluster, these start with KUBERNETES_SERVICE, otherwise look
        # for envs prefixed with KUBECONFIG.
        for key in os.environ:
            if key.startswith("KUBECONFIG") or key.startswith("KUBERNETES_SERVICE"):
                kwargs["env"][key] = os.environ[key]
        kwargs = await super().pre_launch(**kwargs)
        # Determine pod name and namespace - creating the latter if necessary
        self.kernel_pod_name = self._determine_kernel_pod_name(**kwargs)
        self.kernel_namespace = self._determine_kernel_namespace(**kwargs)
        return kwargs

    @overrides
    async def get_provisioner_info(self) -> Dict[str, Any]:
        provisioner_info = await super().get_provisioner_info()
        provisioner_info.update(
            {
                "kernel_ns": self.kernel_namespace,
                "delete_ns": self.delete_kernel_namespace,
            }
        )
        return provisioner_info

    @overrides
    async def load_provisioner_info(self, provisioner_info: dict) -> None:
        await super().load_provisioner_info(provisioner_info)
        self.kernel_namespace = provisioner_info["kernel_ns"]
        self.delete_kernel_namespace = provisioner_info["delete_ns"]

    @overrides
    def get_initial_states(self) -> Set[str]:
        return {"pending", "running"}

    @overrides
    def get_error_states(self) -> Set[str]:
        return {"failed"}

    @overrides
    def get_container_status(self, iteration: Optional[str]) -> str:
        # Locates the kernel pod using the kernel_id selector.  Note that we also include 'component=kernel'
        # in the selector so that executor pods (when Spark is in use) are not considered.
        # If the phase indicates Running, the pod's IP is used for the assigned_ip.
        pod_status = ""
        kernel_label_selector = f"kernel_id={self.kernel_id},component=kernel"
        ret = client.CoreV1Api().list_namespaced_pod(
            namespace=self.kernel_namespace, label_selector=kernel_label_selector
        )
        if ret and ret.items:
            pod_info = ret.items[0]
            self.container_name = pod_info.metadata.name
            if pod_info.status:
                pod_status = pod_info.status.phase.lower()
                if pod_status == "running" and not self.assigned_host:
                    # Pod is running, capture IP
                    self.assigned_ip = pod_info.status.pod_ip
                    self.assigned_host = self.container_name
                    self.assigned_node_ip = pod_info.status.host_ip

        if iteration:  # only log if iteration is not None (otherwise poll() is too noisy)
            self.log.debug(
                f"{iteration}: Waiting to connect to k8s pod in namespace "
                f"'{self.kernel_namespace}'. Name: '{self.container_name}', "
                f"Status: '{pod_status}', Pod IP: '{self.assigned_ip}', "
                f"KernelID: '{self.kernel_id}'"
            )

        return pod_status

    def delete_managed_object(self, termination_stati: List[str]) -> bool:
        """Deletes the object managed by this provisioner

        A return value of True indicates the object is considered deleted,
        otherwise a False or None value is returned.

        Note: the caller is responsible for handling exceptions.
        """
        body = client.V1DeleteOptions(grace_period_seconds=0, propagation_policy="Background")

        # Deleting a Pod will return a v1.Pod if found and its status will be a PodStatus containing
        # a phase string property
        # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.21/#podstatus-v1-core
        v1_pod = client.CoreV1Api().delete_namespaced_pod(
            namespace=self.kernel_namespace, body=body, name=self.container_name
        )
        status = None
        if v1_pod and v1_pod.status:
            status = v1_pod.status.phase

        result = status in termination_stati

        return result

    @overrides
    def terminate_container_resources(self, restart: bool = False) -> Optional[bool]:
        # Kubernetes objects don't go away on their own - so we need to tear down the namespace
        # and/or pod associated with the kernel.  We'll always target the pod first so that shutdown
        # is perceived as happening more rapidly.  Then, if we created the namespace, and we're not
        # in the process of restarting the kernel, we'll delete the namespace.
        # After deleting the pod we check the container status, rather than the status returned
        # from the pod deletion API, since it's not necessarily reflective of the actual status.

        self.restarting = restart
        result = False
        termination_stati = ["Succeeded", "Failed", "Terminating", "Success"]

        # Delete the managed object then, if applicable, the namespace
        object_type = self.object_kind
        try:
            result = self.delete_managed_object(termination_stati)
            if not result:
                # If the status indicates the object is not terminated, capture its current status.
                # If None, update the result to True, else issue warning that it is not YET deleted
                # since we still have the hard termination sequence to occur.
                cur_status = self.get_container_status(None)
                if cur_status is None:
                    result = True
                else:
                    self.log.warning(
                        f"{object_type} '{self.kernel_namespace}.{self.container_name}'"
                        f" is not yet deleted.  Current status is '{cur_status}'."
                    )

            if self.delete_kernel_namespace and restart is False:
                object_type = "Namespace"
                # Status is a return value for calls that don't return other objects.
                # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.21/#status-v1-meta
                body = client.V1DeleteOptions(
                    grace_period_seconds=0, propagation_policy="Background"
                )
                v1_status = client.CoreV1Api().delete_namespace(
                    name=self.kernel_namespace, body=body
                )
                status = None
                if v1_status:
                    status = v1_status.status

                if status and any(s in status for s in termination_stati):
                    result = True

                if not result:
                    self.log.warning(
                        f"Namespace {self.kernel_namespace} is not yet deleted.  "
                        f"Current status is '{status}'."
                    )

        except Exception as err:
            if isinstance(err, client.rest.ApiException) and err.status == 404:
                result = True  # okay if it's not found
            else:
                self.log.warning(f"Error occurred deleting {object_type.lower()}: {err}")

        if result:
            self.log.debug(
                f"KubernetesProcessProxy.terminate_container_resources, "
                f"{self.object_kind}: {self.kernel_namespace}.{self.container_name}, "
                f"kernel ID: {self.kernel_id} has been terminated."
            )
            self.container_name = None
            result = None  # maintain jupyter contract
        else:
            self.log.warning(
                "KubernetesProcessProxy.terminate_container_resources, "
                f"{self.object_kind}: {self.kernel_namespace}.{self.container_name}, "
                f"kernel ID: {self.kernel_id} has not been terminated."
            )

        # Check if there's a kernel pod template file for this kernel and silently delete it.
        kpt_file = kpt_dir + "/kpt_" + self.kernel_id
        try:
            os.remove(kpt_file)
        except OSError:
            pass

        return result

    def _determine_kernel_pod_name(self, **kwargs):
        pod_name = kwargs["env"].get("KERNEL_POD_NAME")
        if pod_name is None:
            pod_name = f"{self.kernel_username}-{self.kernel_id}"

        # Rewrite pod_name to be compatible with DNS name convention
        # And put back into env since kernel needs this
        pod_name = re.sub(r"[^\da-z]+", "-", pod_name.lower())
        while pod_name.startswith("-"):
            pod_name = pod_name[1:]
        while pod_name.endswith("-"):
            pod_name = pod_name[:-1]
        kwargs["env"]["KERNEL_POD_NAME"] = pod_name

        return pod_name

    def _determine_kernel_namespace(self, **kwargs):
        # Since we need the service account name regardless of whether we're creating the namespace or not,
        # get it now.
        service_account_name = KubernetesProvisioner._determine_kernel_service_account_name(
            **kwargs
        )

        # If KERNEL_NAMESPACE was provided, then we assume it already exists.  If not provided, then we'll
        # create the namespace and record that we'll want to delete it as well.
        namespace = kwargs["env"].get("KERNEL_NAMESPACE")
        if namespace is None:
            # check if share gateway namespace is configured...
            if shared_namespace:  # if so, set to RP namespace
                namespace = k8s_provisioner_namespace
                self.log.warning(
                    f"Shared namespace has been configured.  All kernels will reside "
                    f"in the namespace: {namespace}"
                )
            else:
                namespace = self._create_kernel_namespace(service_account_name)
            kwargs["env"]["KERNEL_NAMESPACE"] = namespace  # record in env since kernel needs this
        else:
            self.log.info(f"KERNEL_NAMESPACE provided by client: {namespace}")

        return namespace

    @staticmethod
    def _determine_kernel_service_account_name(**kwargs):
        # Check if an account name was provided.  If not, set to the default name (which can be set
        # from the EG env as well).  Finally, ensure the env value is set.
        service_account_name = kwargs["env"].get(
            "KERNEL_SERVICE_ACCOUNT_NAME", default_kernel_service_account_name
        )
        kwargs["env"]["KERNEL_SERVICE_ACCOUNT_NAME"] = service_account_name
        return service_account_name

    def _create_kernel_namespace(self, service_account_name):
        # Creates the namespace for the kernel based on the kernel username and kernel id.  Since we're creating
        # the namespace, we'll also note that it should be deleted as well.  In addition, the kernel pod may need
        # to list/create other pods (true for spark-on-k8s), so we'll also create a RoleBinding associated with
        # the namespace's default ServiceAccount.  Since this is always done when creating a namespace, we can
        # delete the RoleBinding when deleting the namespace (no need to record that via another member variable).

        namespace = self.kernel_pod_name

        # create the namespace ...
        labels = {"app": app_name, "component": "kernel", "kernel_id": self.kernel_id}
        namespace_metadata = client.V1ObjectMeta(name=namespace, labels=labels)
        body = client.V1Namespace(metadata=namespace_metadata)

        # create the namespace
        try:
            client.CoreV1Api().create_namespace(body=body)
            self.delete_kernel_namespace = True
            self.log.info(f"Created kernel namespace: {namespace}")

            # Now create a RoleBinding for this namespace for the default ServiceAccount.  We'll reference
            # the ClusterRole, but that will only be applied for this namespace.  This prevents the need for
            # creating a role each time.
            self._create_role_binding(namespace, service_account_name)
        except Exception as err:
            # Tolerate pre-existing namespace if we're in the process of restarting
            if isinstance(err, ApiException) and err.status == 409 and self.restarting:
                self.restarting = False
                self.delete_kernel_namespace = True  # still mark for delete
                self.log.info(f"Re-using kernel namespace: {namespace}")
            else:
                if self.delete_kernel_namespace:
                    reason = (
                        f"Error occurred creating role binding for namespace '{namespace}': {err}"
                    )
                    # delete the namespace since we'll be using the EG namespace...
                    body = client.V1DeleteOptions(
                        grace_period_seconds=0, propagation_policy="Background"
                    )
                    client.CoreV1Api().delete_namespace(name=namespace, body=body)
                    self.log.warning(f"Deleted kernel namespace: {namespace}")
                else:
                    reason = f"Error occurred creating namespace '{namespace}': {err}"
                self.log_and_raise(RuntimeError(reason), chained=err)

        return namespace

    def _create_role_binding(self, namespace, service_account_name):
        # Creates RoleBinding instance for the given namespace.  The role used will be the ClusterRole named by
        # EG_KERNEL_CLUSTER_ROLE.
        # Note that roles referenced in RoleBindings are scoped to the namespace so re-using the cluster role prevents
        # the need for creating a new role with each kernel.
        # The ClusterRole will be bound to the kernel service user identified by KERNEL_SERVICE_ACCOUNT_NAME then
        # EG_DEFAULT_KERNEL_SERVICE_ACCOUNT_NAME, respectively.
        # We will not use a try/except clause here since _create_kernel_namespace will handle exceptions.

        role_binding_name = kernel_cluster_role  # use same name for binding as cluster role
        labels = {"app": app_name, "component": "kernel", "kernel_id": self.kernel_id}
        binding_metadata = client.V1ObjectMeta(name=role_binding_name, labels=labels)
        binding_role_ref = client.V1RoleRef(
            api_group="", kind="ClusterRole", name=kernel_cluster_role
        )
        binding_subjects = client.V1Subject(
            api_group="",
            kind="ServiceAccount",
            name=service_account_name,
            namespace=namespace,
        )

        body = client.V1RoleBinding(
            kind="RoleBinding",
            metadata=binding_metadata,
            role_ref=binding_role_ref,
            subjects=[binding_subjects],
        )

        client.RbacAuthorizationV1Api().create_namespaced_role_binding(
            namespace=namespace, body=body
        )
        self.log.info(
            f"Created kernel role-binding '{role_binding_name}' in namespace: {namespace} "
            f"for service account: {service_account_name}"
        )



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Code related to managing kernels running in containers."""
import os
import signal
from abc import abstractmethod
from typing import Any, Dict, List, Optional, Set

import urllib3  # docker ends up using this and it causes lots of noise, so turn off warnings
from jupyter_client import localinterfaces
from overrides import overrides
from traitlets import Unicode, default

from .remote_provisioner import RemoteProvisionerBase

urllib3.disable_warnings()

local_ip = localinterfaces.public_ips()[0]

default_kernel_uid = "1000"  # jovyan user is the default
default_kernel_gid = "100"  # users group is the default

# These could be enforced via a PodSecurityPolicy, but those affect
# all pods so the cluster admin would need to configure those for
# all applications.
prohibited_uids = os.getenv("GP_PROHIBITED_UIDS", "0").split(",")
prohibited_gids = os.getenv("GP_PROHIBITED_GIDS", "0").split(",")

mirror_working_dirs = bool(os.getenv("GP_MIRROR_WORKING_DIRS", "false").lower() == "true")


class ContainerProvisionerBase(RemoteProvisionerBase):
    """Kernel provisioner for container-based kernels."""

    image_name_env = "GP_IMAGE_NAME"
    image_name = Unicode(
        None,
        config=True,
        allow_none=True,
        help="""The image name to use when launching container-based kernels.
(GP_IMAGE_NAME env var)""",
    )

    @default("image_name")
    def _image_name_default(self):
        return os.getenv(self.image_name_env)

    executor_image_name_env = "GP_EXECUTOR_IMAGE_NAME"
    executor_image_name = Unicode(
        None,
        config=True,
        allow_none=True,
        help="""The image name to use as the Spark executor image when launching
container-based kernels within Spark environments. (GP_EXECUTOR_IMAGE_NAME env var)""",
    )

    @default("executor_image_name")
    def _executor_image_name_default(self):
        return os.getenv(self.executor_image_name_env) or self.image_name

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.container_name = None
        self.assigned_node_ip = None

    @property
    @overrides
    def has_process(self) -> bool:
        return self.container_name is not None

    @overrides
    async def pre_launch(self, **kwargs: Any) -> Dict[str, Any]:
        # Unset assigned_host, ip, and node_ip in pre-launch, otherwise, these screw up restarts
        self.assigned_host = ""
        self.assigned_ip = None
        self.assigned_node_ip = None

        kwargs = await super().pre_launch(**kwargs)

        kwargs["env"]["KERNEL_IMAGE"] = self.image_name
        kwargs["env"]["KERNEL_EXECUTOR_IMAGE"] = self.executor_image_name

        if (
            not mirror_working_dirs
        ):  # If mirroring is not enabled, remove working directory if present
            if "KERNEL_WORKING_DIR" in kwargs["env"]:
                del kwargs["env"]["KERNEL_WORKING_DIR"]

        self._enforce_prohibited_ids(**kwargs)
        return kwargs

    @overrides
    def log_kernel_launch(self, cmd: List[str]) -> None:
        self.log.info(
            f"{self.__class__.__name__}: kernel launched. Kernel image: {self.image_name}, "
            f"KernelID: {self.kernel_id}, cmd: '{cmd}'"
        )

    def _enforce_prohibited_ids(self, **kwargs):
        """Determine UID and GID with which to launch container and ensure they are not prohibited."""
        kernel_uid = kwargs["env"].get("KERNEL_UID", default_kernel_uid)
        kernel_gid = kwargs["env"].get("KERNEL_GID", default_kernel_gid)

        if kernel_uid in prohibited_uids:
            error_message = (
                f"Kernel's UID value of '{kernel_uid}' has been denied via GP_PROHIBITED_UIDS!"
            )
            self.log_and_raise(PermissionError(error_message))
        elif kernel_gid in prohibited_gids:
            error_message = (
                f"Kernel's GID value of '{kernel_gid}' has been denied via GP_PROHIBITED_GIDS!"
            )
            self.log_and_raise(PermissionError(error_message))

        # Ensure the kernel's env has what it needs in case they came from defaults
        kwargs["env"]["KERNEL_UID"] = kernel_uid
        kwargs["env"]["KERNEL_GID"] = kernel_gid

    @overrides
    async def poll(self) -> Optional[int]:
        """Determines if container is still active.

        Submitting a new kernel to the container manager will take a while to be Running.
        Thus, kernel ID will probably not be available immediately for poll.
        So will regard the container as active when no status is available or one of the initial
        phases.
        Returns None if the container cannot be found or its in an initial state. Otherwise,
        return an exit code of 0.
        """
        result = 0

        container_status = self.get_container_status(None)
        # Do not check whether container_status is None
        # EG couldn't restart kernels although connections exists.
        # See https://github.com/jupyter/enterprise_gateway/issues/827
        if container_status in self.get_initial_states():
            result = None

        return result

    @overrides
    async def send_signal(self, signum: int) -> None:
        """Send signal `signum` to container."""
        if signum == 0:
            return await self.poll()
        elif signum == signal.SIGKILL:
            return await self.kill()
        else:
            # This is very likely an interrupt signal, so defer to the super class
            # which should use the communication port.
            return await super().send_signal(signum)

    @overrides
    async def kill(self, restart: bool = False) -> None:
        """Kills a containerized kernel."""
        result = None

        if self.container_name:  # We only have something to terminate if we have a name
            result = self.terminate_container_resources(restart=restart)

        return result

    @overrides
    async def terminate(self, restart: bool = False) -> None:
        """Terminates a containerized kernel.

        This method defers to kill() since there's no distinction between the
        two in these environments.
        """
        return await self.kill(restart=restart)

    @overrides
    async def shutdown_listener(self, restart: bool) -> None:
        await super().shutdown_listener(restart)
        if self.container_name:  # We only have something to terminate if we have a name
            self.terminate_container_resources(restart)

    @overrides
    async def confirm_remote_startup(self):
        """Confirms the container has started and returned necessary connection information."""
        self.log.debug("Trying to confirm kernel container startup status")
        self.start_time = RemoteProvisionerBase.get_current_time()
        i = 0
        ready_to_connect = False  # we're ready to connect when we have a connection file to use
        while not ready_to_connect:
            i += 1
            await self.handle_launch_timeout()

            container_status = self.get_container_status(str(i))
            if container_status:
                if container_status in self.get_error_states():
                    reason = f"Error starting kernel container; status: '{container_status}'.  Check server logs."
                    self.log_and_raise(RuntimeError(reason))
                else:
                    if self.assigned_host != "":
                        ready_to_connect = await self.receive_connection_info()
                        self.pid = (
                            0  # We won't send the process signals from container-based provisioners
                        )
                        self.pgid = 0
            else:
                self.detect_launch_failure()

    @overrides
    async def get_provisioner_info(self) -> Dict[str, Any]:
        """Captures the base information necessary for kernel persistence relative to containers."""
        provisioner_info = await super().get_provisioner_info()
        provisioner_info.update(
            {
                "assigned_node_ip": self.assigned_node_ip,
            }
        )
        return provisioner_info

    @overrides
    async def load_provisioner_info(self, provisioner_info: dict) -> None:
        """Loads the base information necessary for kernel persistence relative to containers."""
        await super().load_provisioner_info(provisioner_info)
        self.assigned_node_ip = provisioner_info.get("assigned_node_ip")

    @abstractmethod
    def get_initial_states(self) -> Set[str]:
        """Return list of states (in lowercase) indicating container is starting (includes running)."""
        raise NotImplementedError

    @abstractmethod
    def get_error_states(self) -> Set[str]:
        """Returns the list of error states (in lowercase)."""
        raise NotImplementedError

    @abstractmethod
    def get_container_status(self, iteration: Optional[str]) -> str:
        """Return current container state."""
        raise NotImplementedError

    @abstractmethod
    def terminate_container_resources(self, restart: bool = False) -> Optional[bool]:
        """Terminate any artifacts created on behalf of the container's lifetime."""
        raise NotImplementedError
